//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Boundary.cpp: implementation of the Boundary class.
//
//////////////////////////////////////////////////////////////////////

#include "Boundary.h"
#include <maya/MDagPath.h>
#include <maya/MFnMesh.h>
#include <maya/MFnNurbsSurface.h>
#include "MELUtilities.h"
#include <sstream>
#include <iostream>
#include <cmath>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Boundary::Boundary()
{
	
}

Boundary::Boundary(wallBehaviour wallType, double wc, int we)
{
	_nPolySurfs = 0;
	_nNurbsSurfs = 0;
	_wallBehaviour = wallType;
	_const = wc;
	_exponent = we;
	_threshold = setThreshold(0.01);
}

Boundary::~Boundary()
{

}


//	Parameters:
//		Vector pos
//			The starting position of the point to be moved.
//		Vector step
//			The movement of the point (movement is a straight line).
//	Return:
//		The correction to the movement
//	Other:
//		This function is called by the vertex as it tries to move.
//		It checks all the surfaces for intersection and calculates a correction to the step that is returned.

Vector Boundary::getWallDisplacement(Vector pos, Vector step)
{
	double minDist = MIN_DIST, minDistX = MIN_DIST, minDistY = MIN_DIST, minDistZ = MIN_DIST, differ;
	Vector disp, intersect, diff, xintersect, yintersect, zintersect;
	bool inter;
	//Check the polysurfs
	for(int i=0; i< _nPolySurfs; i++)	{
		if(_wallBehaviour==CUT_OFF)	{
			inter = _polySurfs[i]->getIntersection(pos, step, intersect);
			if(inter)	{
				diff = pos - intersect;
				if(wallTooClose(diff.getLength(), minDist, step.getLength()))	{		
					disp = getWallForce(pos, step, intersect);
					minDist = diff.getLength();
				}
			}
		}
		else	{
			inter = _polySurfs[i]->getIntersection(pos, step, xintersect, yintersect, zintersect);
			if(inter)	{
				//cout << "Found intersection with wall " << _polySurfs[i]->getID() << endl;
				differ = fabs(pos.getX() - xintersect.getX());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistX, fabs(step.getX())))	{		
					disp.setX(getWallForce(pos.getX(), step.getX(), xintersect.getX()));
					minDistX = differ;
					//cout << "The wall is at distance " << minDistX << endl;
				}
				differ = fabs(pos.getY() - yintersect.getY());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistY, fabs(step.getY())))	{		
					disp.setY(getWallForce(pos.getY(), step.getY(), yintersect.getY()));
					minDistY = differ;
					//cout << "The wall is at distance " << minDistY << endl;
				}
				differ = fabs(pos.getZ() - zintersect.getZ());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistZ, fabs(step.getZ())))	{		
					disp.setZ(getWallForce(pos.getZ(), step.getZ(), zintersect.getZ()));
					minDistZ = differ;
					//cout << "The wall is at distance " << minDistZ << endl;
				}
			}
		}
	}
	//Check the nurbs
	for(int j=0; j< _nurbsSurfs.size(); j++)	{
		if(_wallBehaviour==CUT_OFF)	{
			inter = _nurbsSurfs[j]->getIntersection(pos, step, intersect);
			if(inter)	{
				cout << "Found intersection with wall " << _nurbsSurfs[j]->getID() << endl;
				diff = pos - intersect;
				//cout << "The vector from the wall to the vertex is " << diff.printVector().c_str() << endl;
				if(wallTooClose(diff.getLength(), minDist, step.getLength()))	{		
					disp = getWallForce(pos, step, intersect);
					minDist = diff.getLength();
					//cout << "The wall is at distance " << minDist << endl;
				}
			}
		}
		else	{
			inter = _nurbsSurfs[j]->getIntersection(pos, step, xintersect, yintersect, zintersect);
			if(inter)	{
				//cout << "Found intersection with wall " << _nurbsSurfs[j]->getID() << endl;
				differ = fabs(pos.getX() - xintersect.getX());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistX, fabs(step.getX())))	{		
					disp.setX(getWallForce(pos.getX(), step.getX(), xintersect.getX()));
					minDistX = differ;
					//cout << "The wall is at distance " << minDistX << endl;
				}
				differ = fabs(pos.getY() - yintersect.getY());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistY, fabs(step.getY())))	{		
					disp.setY(getWallForce(pos.getY(), step.getY(), yintersect.getY()));
					minDistY = differ;
					//cout << "The wall is at distance " << minDistY << endl;
				}
				differ = fabs(pos.getZ() - zintersect.getZ());
				//cout << "The vector from the wall to the vertex is " << differ << endl;
				if(wallTooClose(differ, minDistZ, fabs(step.getZ())))	{		
					disp.setZ(getWallForce(pos.getZ(), step.getZ(), zintersect.getZ()));
					minDistZ = differ;
					//cout << "The wall is at distance " << minDistZ << endl;
				}
			}
		}
	}
	//cout << "The wall displacement is " << disp.printVector().c_str() << endl;
	return disp;
}

//	Parameters:
//		Vector pos
//			The starting position of the point to be moved.
//		Vector step
//			The movement.
//	Return:
//		True if the point passed through a wall as it moved.
//	Other:
//		This is used by the AgentSoftBoundary to see how many vertices violates the boundary conditions

bool Boundary::passedWall(Vector pos, Vector step)
{
	//Just check intersections with each wall
	Vector intersect;
	bool inter;
	for(int i=0; i< _nPolySurfs; i++)	{
		inter = _polySurfs[i]->getIntersection(pos, step, intersect);
		if(inter)
			return true;
	}
	for(int j=0; j< _nurbsSurfs.size(); j++)	{
		inter = _nurbsSurfs[j]->getIntersection(pos, step, intersect);
		if(inter)
			return true;
	}
	return false;
}

//	Parameters:
//		Vector pos
//			The starting pos for the moving point
//		Vector step
//			The movement
//		Vector intersect
//			The point of intersection with a surface
//	Return:
//		The correction due to the wall
//	Other:
//		This is a vector version that is used for the cut-off behaviour

Vector Boundary::getWallForce(Vector pos, Vector step, Vector intersect)
{
	Vector disp, partialstep, tmp, diff = pos - intersect;
	double dist = diff.getLength();
	if(_wallBehaviour==ACTIVE_WALLS)	{
		//The wall acts as a repellor
		disp = step;
		disp.normalize();
		disp = disp*magnitude(dist, _const, _exponent, true);
		//Check if force to big, the wall can not push the vertex back
		disp = disp.getLength() > step.getLength() ? step : disp;
	}
	else	{
		disp = step + diff;
	}
	//We want to have a little distance to the wall, that is why we use the tol
	return disp*(1 + 3*TOL);
}

//This is the same as the above except that it has scalar arguments. It is used for the default
//wall behaviour when we check each component separately.

double Boundary::getWallForce(double pos, double step, double intersect)
{
	return (step + pos - intersect)*(1 + 3*TOL);
}

//	Parameters:
//		double tol
//			The level of tolerance
//	Return:
//		What this means for wall distances
//	Other:

double Boundary::setThreshold(double tol)
{
	return pow(_const/tol, _exponent);
}

//	Parameters:
//		double dist
//			The distance to the wall.
//		double closeset
//			Distance to the closest wall so far.
//		double step
//			How far we are trying to move
//	Return:
//		True if the wall is so close that we need to adjust the step
//	Other:
//		For the ACTIVE_WALLS we have threshold value after which the adjustments are so small that we do not bother.

bool Boundary::wallTooClose(double dist, double closest, double step)
{
	if(_wallBehaviour==ACTIVE_WALLS)
		return (dist < closest) && (dist < _threshold);
	else
		return (dist < closest) && (dist < step);
}

//	Parameters:
//		MDagPath surf
//			The Maya object.
//	Return:
//	Other:
//		Use this when constructing surfaces

void Boundary::addNurbsSurface(MDagPath surf)
{
	NurbsSurface* nurbsSurf = new NurbsSurface(surf, _nNurbsSurfs++);
	_nurbsSurfs.push_back(nurbsSurf);
}

void Boundary::addPolySurface(MDagPath surf)
{
	PolygonSurface* polySurf = new PolygonSurface(surf, _nPolySurfs++);
	_polySurfs.push_back(polySurf);
}