//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// LSystem.cpp: implementation of the LSystem class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <sstream>
#include "LSystem.h"
#include "BNF.h"
#include "RewriteRule.h"
#include "Axiom.h"
#include "MELUtilities.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LSystem::LSystem()
{

}

LSystem::LSystem(BNF* grammar, BNFtype bnfType)
: NonTerminal(grammar, bnfType)
{
	_name = L_SYSTEM;
}

LSystem::~LSystem()
{

}

Sentence* LSystem::getProductionRule(int gene, bool terminal)
{
	ostringstream str;
	str << AXIOM << " " << REWRITE_RULE << " { " << REWRITE_RULE << " }";
	return new Sentence(str.str());
}

//Used by regn8

void LSystem::reconstruct(Sentence* sent, Genotype* &genome, int multi)
{
	int i = 0;
	vector<RewriteRule*> rewriteRules;
	vector<Sentence*> lines;
	//Since we only have one production rule, it does not matter what gene we have
	genome->addRandGene();
	//Parse the sentence
	//First line is the Axiom
	Sentence *line, *axLine = sent->getLine(0);
	Axiom* ax = new Axiom;
	//Then comes the rewrite rules
	line = sent->getLine(++i);
	line->printSentence();
	RewriteRule* rewrite = new RewriteRule;
	rewriteRules.push_back(rewrite);
	lines.push_back(line);
	line = sent->getLine(++i);
	//The optional rules
	while(line->isRewriteRule())	{
		rewrite = new RewriteRule;
		genome->addGene(i-1);
		rewriteRules.push_back(rewrite);
		lines.push_back(line);
		line = sent->getLine(++i);
		//line->printSentence();
	}
	//Make sure that we do not get any more edges
	genome->addGene(0);
	ax->reconstruct(axLine, genome);
	for(int j=0; j<i-1; j++)	{
		rewriteRules[j]->reconstruct(lines[j], genome);
	}
}

//Overloading of the non-terminal function.
//Since we want to include the repair mechanism

void LSystem::expand(int depth, int predecessor)
{
	depth--;
	bool multi = false, optional = false;
	int gene, index, nWords;
	_BNF->getNextGene(gene, index);
	//Parse the productionrule
	Sentence* prodRule = getProductionRule(gene, depth<=0);
	string term;
	//cout << "Expanding a " << _name.c_str() << " node with production rule 0" << endl;
	nWords = prodRule->numberOfWords();
	for(int i=0; i<nWords; i++)	{
		term = prodRule->getWord(i)->getName();
		//cout << "Found term " << tmpTerm << endl;
		if(term==AXIOM)
			addAxiom();
		else if(term==REWRITE_RULE)
			NonTerminal::addRewriteRule(multi);
		//Special characters
		else if(term=="{")	//Multiple occurences
			multi = true;
		else if(term=="}")	//Multiple occurences
			multi = false;
		else if(term=="|")	//Optional word
			optional = !optional;
		//Special end-symbol
		else if(term==END_VAR)	{
			//Vegetate
		}
	}
	//Expand the children
	//cout << "Sentence consists of " << _sentence->numberOfWords() << " words." << endl;
	//_sentence->printSentence();
	string name;
	for(int j=0; j<_sentence->numberOfWords(); j++)	{
		name = _sentence->getWord(j)->getName();
		if(name==AXIOM || name==L_SYSTEM || name==REWRITE_RULE || name==PREDECESSOR ||
			name==SUCCESSOR || name==MODIFIER || name==OPERATOR || name==CONDITION)
			((NonTerminal*)_sentence->getWord(j))->expand(depth);
	}
	completeRewriteRules(depth);
	_expandable = false;
	//cout << "Finished expanding a " << _name.c_str() << " node with production rule 0" << endl;
}

//This function chekcs the productions and make sure that there is at least one production
//for each edgetype.

void LSystem::completeRewriteRules(int depth)
{
	//cout << "Completing sentence." << endl;
	//Check that all edges are represented in the predecessors
	vector<bool> edges;
	for(int k=0; k<_BNF->getEdgeTypes(); k++)
		edges.push_back(false);

	Sentence *line, *pred, *sent = _BNF->getFullSentence();
	int nLines, arrowpos;
	//Get all the predecessors
	nLines = sent->numberOfLines();
	//cout << "Expanded sentence with " << nLines << " lines: " << endl;
	//sent->printSentence();
	//Parse the sentence
	for(int i=0; i<nLines; i++)	{
		line = sent->getLine(i);
		//cout << "Extracted line:" << endl;
		//line->printSentence();
		arrowpos = line->findWord("->");
		if(arrowpos>0)	{
			//There are only four different predecessors so we check all of them
			pred = line->getSentence(0, arrowpos);
			if(pred->numberOfWords()==1)
				edges[pred->getWord(0)->getEdgeType()] = true;
			else if(pred->numberOfWords()==3)	{
				if(pred->getWord(1)->getName()=="<")
					edges[pred->getWord(2)->getEdgeType()] = true;
				else
					edges[pred->getWord(0)->getEdgeType()] = true;
			}
			else
				edges[pred->getWord(2)->getEdgeType()] = true;
		}
	}
	//Add rewrite rules for the edges that lack those
	for(int l=0; l<_BNF->getEdgeTypes(); l++)	{
		//cout << "For edge " << l << " the entry is " << boolalpha << edges[l] << endl;
		if(!edges[l])
			addRewriteRule(l, depth);
	}
}

//Add a production to the grammar that has a specified type as predecessor.

void LSystem::addRewriteRule(int predecessor, int depth)
{
	//cout << "Adding rewriterule " << predecessor << endl;
	RewriteRule* rewrite = new RewriteRule(_BNF, _BNFtype);
	_sentence->addWord((Word*)rewrite);
	//Word* end = new Word("end");
	//_sentence->addWord(end);
	//Expand it
	//cout << "Expanding " << _sentence->getWord(_sentence->numberOfWords()-2)->getName() << endl;
	((NonTerminal*)_sentence->getWord(_sentence->numberOfWords()-1))->expand(depth, predecessor);
}
