//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// NonTerminal.cpp: implementation of the NonTerminal class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include "NonTerminal.h"
#include "Modifier.h"
#include "Axiom.h"
#include "RewriteRule.h"
#include "Successor.h"
#include "Predecessor.h"
#include "Operator.h"
#include "Edge.h"
#include "OperatorType.h"
#include "BNF.h"
#include "MELUtilities.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NonTerminal::NonTerminal()
{

}

//This constructor is used by all the child classes

NonTerminal::NonTerminal(BNF* grammar, BNFtype bnfType)
:	Terminal()
{
	_BNF = grammar;
	_BNFtype = bnfType;
	_expandable = true;
}

NonTerminal::~NonTerminal()
{

}

//Use this function to expand a non-terminal by choosing a production rule
//When depth<=0, we try to choose a production rule that only results in terminals
//The: int predecessor argument is used when we want to have a specific predecessor
//If predecessor!=-1, we should have that type as a Predecessor and we should not introduce any new
//edge types because we are now in the repair phase.

void NonTerminal::expand(int depth, int predecessor)
{
	depth--;
	int edgeXtype = -1, edgeYtype = -1, edgeZtype = -1, gene, index, nWords;
	bool multi = false, optional = false;
	//cout << "The gene was " << productionRule << " and there are " << _productionRules.size() << " production rules." << endl;
	//Get a gene from the genome. The BNF-class keeps track of where we are reading.
	_BNF->getNextGene(gene, index);
	//Choose which productionrule to use
	Sentence* prodRule = getProductionRule(gene, depth<=0);
	string term;
	//cout << "Expanding a " << _name.c_str() << " node with production rule " << prod << ", depth is " << depth << endl;
	//Parse the productionrule
	nWords = prodRule->numberOfWords();
	for(int i=0; i<nWords; i++)	{
		term = prodRule->getWord(i)->getName();
		//Non-terminals
		if(term==AXIOM)
			addAxiom();
		else if(term==REWRITE_RULE)	{
			addRewriteRule(multi);
		}
		else if(term==OPERATOR)	{
			addOperator(multi);
		}
		else if(term==PREDECESSOR)
			addPredecessor();
		else if(term==SUCCESSOR)
			addSuccessor(multi);
		else if(term==MODIFIER)
			addModifier(multi);
		//Special characters
		else if(term=="{")	//Multiple occurences
			multi = true;
		else if(term=="}")	//Multiple occurences
			multi = false;
		else if(term=="|")	//Optional word
			optional = !optional;
		//Edges (or segments)
		else if(term==EDGE)	{
			//cout << "_name is " << _name.c_str() << " and predecessor is " << predecessor << endl;
			if(_name==PREDECESSOR)	{
				if(predecessor!=-1 && ( (nWords==1 && i==0) || (nWords==3 && i==2) || (nWords==2 && i==0) || (nWords==5 && i==2)))
					addEdge(predecessor);
				else
					addEdge(multi, false);
			}
			else
				addEdge(multi, predecessor==-1);
		}
		//Make sure that all edges generated from this non-terminal have the same type
		else if(term==EDGE_X)	{
			if(edgeXtype==-1)	//First occurence
				addEdge(multi, predecessor==-1, edgeXtype);
			else
				addEdge(edgeXtype);
		}
		else if(term==EDGE_Y)	{
			if(edgeYtype==-1)	//First occurence
				addEdge(multi, predecessor==-1, edgeYtype);
			else
				addEdge(edgeYtype);
		}
		else if(term==EDGE_Z)	{
			if(edgeZtype==-1)	//First occurence
				addEdge(multi, predecessor==-1, edgeZtype);
			else
				addEdge(edgeZtype);
		}
		//Special symbol for the axiom, StartEdge = +Edge
		else if(term==START_EDGE)
			addStartEdge(multi);
		//Special end-symbol
		else if(term==END_VAR)	{
			//Vegetate
		}
		//Only operators left
		else
			addOperatorType(term);
	}
	//Expand the children
	//cout << "Sentence consists of " << _sentence->numberOfWords() << " words." << endl;
	//_sentence->printSentence();
	string name;
	for(int j=0; j<_sentence->numberOfWords(); j++)	{
		name = _sentence->getWord(j)->getName();
		if(name==AXIOM || name==L_SYSTEM || name==REWRITE_RULE || name==PREDECESSOR ||
			name==SUCCESSOR || name==MODIFIER || name==OPERATOR || name==CONDITION)
			((NonTerminal*)_sentence->getWord(j))->expand(depth, predecessor);
	}
	_expandable = false;
	//cout << "Finished expanding a " << _name.c_str() << " node with production rule " << prod << endl;
}


void NonTerminal::addAxiom()
{
	Word* end = new Word(END_VAR);
	Axiom* axiom = new Axiom(_BNF, _BNFtype);
	_sentence->addWord((Word*)axiom);
	_sentence->addWord(end);
}

void NonTerminal::addRewriteRule(bool multi)
{
	if(!multi)	{
		RewriteRule* rewrite = new RewriteRule(_BNF, _BNFtype);
		_sentence->addWord((Word*)rewrite);
	}
	else	{
		int occurences = 0, index, gene;
		_BNF->getNextGene(gene, index);
		//We use the genes to determine how many rewrite rules we should have when the number is optional
		//The probability for adding a new rewrite rules decreases as new ones are added.
		//This scheme gives no upper boundary to the number of rewrite rules, although,
		//we can easily limit the expected value
		int prob = gene % (occurences + 3);
		while( prob>occurences )	{
			RewriteRule* rewrite = new RewriteRule(_BNF, _BNFtype);
			_sentence->addWord((Word*)rewrite);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 3);
		}
	}
}

void NonTerminal::addOperator(bool multi)
{
	if(multi)	{
		int occurences = 0, index, gene;
		_BNF->getNextGene(gene, index);
		int prob =  gene % (occurences + 2);
		while( prob>occurences )	{
			Operator* op = new Operator(_BNF, _BNFtype);
			_sentence->addWord((Word*)op);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		Operator* op = new Operator(_BNF, _BNFtype);
		_sentence->addWord((Word*)op);
	}
}

void NonTerminal::addPredecessor()
{
	Predecessor* pred = new Predecessor(_BNF, _BNFtype);
	_sentence->addWord((Word*)pred);
}

void NonTerminal::addSuccessor(bool multi)
{
	if(multi)	{
		int occurences = 0, index, gene;
		_BNF->getNextGene(gene, index);
		int prob = gene % (occurences + 2);
		while( prob>occurences )	{
			Word* transform = new Word("->");
			_sentence->addWord(transform);
			Successor* succ = new Successor(_BNF, _BNFtype);
			_sentence->addWord((Word*)succ);
			Word* end = new Word(END_VAR);
			_sentence->addWord(end);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		Word* transform = new Word("->");
		_sentence->addWord(transform);
		Successor* succ = new Successor(_BNF, _BNFtype);
		_sentence->addWord((Word*)succ);
		Word* end = new Word(END_VAR);
		_sentence->addWord(end);
	}
}

void NonTerminal::addModifier(bool multi)
{
	if(multi)	{
		int occurences = 0, index, gene;
		_BNF->getNextGene(gene, index);
		int prob = gene % (occurences + 2);
		while( prob>occurences )	{
			Modifier* mod = new Modifier(_BNF, _BNFtype);
			_sentence->addWord((Word*)mod);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		Modifier* mod = new Modifier(_BNF, _BNFtype);
		_sentence->addWord((Word*)mod);
	}
}

//When we are in the reapir phase we do not want to introduce new edge types

void NonTerminal::addEdge(bool multi, bool newEdge)
{
	int ran, index, gene;
	_BNF->getNextGene(gene, index);
	if(multi)	{
		int occurences = 0;
		int prob = gene % (occurences + 2);
		while( prob > occurences)	{
			//Determine the type of the new edge
			int nTypes = _BNF->getEdgeTypes();
			_BNF->getNextGene(gene, index);
			ran = _BNF->getEdge(newEdge);

			Edge* edg = new Edge(ran);
			_sentence->addWord((Word*)edg);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		//Determine the type of the new edge
		ran = _BNF->getEdge(newEdge);
		Edge* edg = new Edge(ran);
		_sentence->addWord((Word*)edg);
	}
}

//Use this when we know what type the edge should have

void NonTerminal::addEdge(int predecessor)
{
	Edge* edg = new Edge(predecessor);
	_sentence->addWord((Word*)edg);
}

//Use this when we need to know what type the edge got.
//It is used by the EdgeX, Y, and Z.
//These are used so that we can get the same edge types for several edges in a production

void NonTerminal::addEdge(bool multi, bool newEdge, int &edgetype)
{
	int ran, index, gene;
	_BNF->getNextGene(gene, index);
	if(multi)	{
		int occurences = 0;
		int prob = gene % (occurences + 2);
		while( prob > occurences)	{
			//Determine the type of the new age
			int nTypes = _BNF->getEdgeTypes();
			if(edgetype!=-1)
				ran = edgetype;
			else	{
				ran = _BNF->getEdge(newEdge);
				edgetype = ran;
			}
			Edge* edg = new Edge(ran);
			_sentence->addWord((Word*)edg);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		//Determine the type of the new age
		edgetype = _BNF->getEdge(newEdge);
		Edge* edg = new Edge(edgetype);
		_sentence->addWord((Word*)edg);
	}
}

//This is just a symbol for "+ Edge"

void NonTerminal::addStartEdge(bool multi)
{
	int gene, index, ran;
	_BNF->getNextGene(gene, index);
	if(multi)	{
		int occurences = 0;
		int prob = gene % (occurences + 2);
		while( prob > occurences )	{
			//First add a '+'
			OperatorType* op = new OperatorType(TURN_RIGHT);
			_sentence->addWord((Word*)op);
			//Determine the type of the new age
			ran = _BNF->getEdge(true);
			Edge* edg = new Edge(ran);
			_sentence->addWord((Word*)edg);
			occurences++;
			_BNF->getNextGene(gene, index);
			prob = gene % (occurences + 2);
		}
	}
	else	{
		//First add a '+'
		OperatorType* op = new OperatorType(TURN_RIGHT);
		_sentence->addWord((Word*)op);
		//Determine the type of the new age
		ran = _BNF->getEdge(true);
		Edge* edg = new Edge(ran);
		_sentence->addWord((Word*)edg);
	}
}

void NonTerminal::addOperatorType(string oper)
{
	OperatorType* op;
	if(oper=="-")
		op = new OperatorType(TURN_LEFT);
	else if(oper=="+")
		op = new OperatorType(TURN_RIGHT);
	else if(oper=="&")
		op = new OperatorType(PITCH_DOWN);
	else if(oper=="^")
		op = new OperatorType(PITCH_UP);
	else if(oper=="\\")
		op = new OperatorType(ROLL_LEFT);
	else if(oper=="/")
		op = new OperatorType(ROLL_RIGHT);
	else if(oper=="~")
		op = new OperatorType(CHANGE_DIRECTION);
	else if(oper=="[")
		op = new OperatorType(PUSH);
	else if(oper=="]")
		op = new OperatorType(POP);
	else if(oper=="<")
		op = new OperatorType(LEFT_SENSITIVE);
	else if(oper==">")
		op = new OperatorType(RIGHT_SENSITIVE);
	else if(oper=="->")
		op = new OperatorType(TRANSFORM);
		
	_sentence->addWord((Word*)op);
}

void NonTerminal::printNonTerminal()
{
	string name;
	for(int i=0; i<_sentence->numberOfWords(); i++)	{
		name = _sentence->getWord(i)->getName();
		if(name==AXIOM || name==L_SYSTEM || name==REWRITE_RULE || name==PREDECESSOR ||
			name==SUCCESSOR || name==MODIFIER || name==OPERATOR || name==CONDITION)
			((NonTerminal*)_sentence->getWord(i))->printNonTerminal();
		else if(name==END_VAR)
			cout << endl;
		else
			_sentence->getWord(i)->printWord();
	}
}

