//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Region.cpp: implementation of the Region class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <sstream>
#include "Region.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Region::Region()
{
	_nSegments = 0;
	_nVertices = 0;
}

Region::Region(int ID)
{
	_ID = ID;
	_nSegments = 0;
	_nVertices = 0;
}

Region::~Region()
{

}

//Just add the segment to the array, we do not care about order for now

void Region::addSegment(Segment* &segment)
{
	_segments.push_back(segment);
	_nSegments++;
	//The vertices are automatically adjacent to the region
	if(segment->getEnd()!=NULL)
		addVertex(segment->getEnd());
	if(segment->getStart()!=NULL)
		addVertex(segment->getStart());
}

bool Region::surroundedByVertex(Vertex* ver)
{
	//cout << "There are " << _nVertices << " or " << _vertices.size() << " vertices surrounding region " << _ID << endl;
	for(int i=0; i<_nVertices; i++)	{
		//cout << "Found that vertex " << _vertices[i]->getID() << " surrounds region." << endl;
		if(_vertices[i]->getID()==ver->getID())
			return true;
	}
	return false;
}

//Just add the vertex, we do not care about the order now.
//This function is called by the addSegment

void Region::addVertex(Vertex* vertex)
{
	//Check that it does not already exist in the array
	if(!surroundedByVertex(vertex))	{
		_vertices.push_back(vertex);
		_nVertices++;
	}
}

//This function is used when the region is split

void Region::removeVertex(Vertex* vertex)
{
	//cout << "Removing vertex " << vertex->getID() << " from region " << _ID << endl;
	vector<Vertex*>::iterator iter=_vertices.begin();
	int i = 0;
	while(iter != _vertices.end())	{
		if(_vertices[i]->getID()==vertex->getID())	{
			_vertices.erase(iter);
			_nVertices--;
			return;
		}
		iter++;
		i++;
	}
	//cout << "Failed to remove vertex." << endl;
}

//Parameters:
//	Vertex *start
//		The vertex where the new segment starts
//	Vertex *end
//		The vertex where the new segment ends
//	vector<Segment*> &segmentsForNewRregion
//		An array of the segments that will be surround the new region
//Return:
//Note:
//	This is called from newRegion in Scaffold when a new region is created.

void Region::splitRegion(Vertex *start, Vertex *end, vector<Segment*> &segmentsForNewRegion/*, vector<Vertex*> &verticesForNewRegion*/)
{
	segmentsForNewRegion.clear();
	int i = 0;
	Vertex* ver;
	Segment* seg;
	bool startOfSeg;
	vector<Segment*>::iterator it = _segments.begin();
	//Find the starting vertex
	while(_segments[i]->getStart()->getID()!=start->getID() &&
			_segments[i]->getEnd()->getID()!=start->getID() &&
			i < _nSegments)	{
		it++;
		i++;
	}

	startOfSeg = _segments[i]->getStart()->getID()==start->getID();
	//Add segments that should be used for new region to array and remove them
	segmentsForNewRegion.push_back(_segments[i]);
	//cout << "Segment " << _segments[i]->getID() << " is going to surround the new region" << endl;
	//Find the direction to move in
	if(startOfSeg)
		ver = _segments[i]->getEnd();
	else
		ver = _segments[i]->getStart();
	seg = _segments[i];
	//cout << "Removing segment " << seg->getID() << " from region " << _ID << endl;
	_segments.erase(it);
	removeVertex(ver);
	ver->removeRegion(this);
	do	{
		getNextSegment(seg, ver, it);
		segmentsForNewRegion.push_back(seg);
		//cout << "Segment " << seg->getID() << " is going to surround the new region" << endl;
		//cout << "Removing segment " << seg->getID() << " from region " << _ID << endl;
		_segments.erase(it);
		ver->removeRegion(this);
		removeVertex(ver);
	}
	while(ver->getID()!=end->getID());
}
		
void Region::getNextSegment(Segment* &seg, Vertex* &ver)
{
	//cout << "Looking for the segment at the other side of vertex " << ver->getID() << " from segment " << seg->getID() << endl;
	for(int i=0; i<_nSegments; i++)	{
		//cout << "Segment " << _segments[i]->getID() << " starts at " << _segments[i]->getStart()->getID() << " and ends at " << _segments[i]->getEnd()->getID() << endl;
		if(_segments[i]->getStart()->getID()==ver->getID() && _segments[i]->getID()!=seg->getID() )	{
			seg = _segments[i];
			ver = seg->getEnd();
			return;
		}
		else if(_segments[i]->getEnd()->getID()==ver->getID() && _segments[i]->getID()!=seg->getID())	{
			seg = _segments[i];
			ver = seg->getStart();
			return;
		}
	}
}

void Region::getNextSegment(Segment* &seg, Vertex* &ver, vector<Segment*>::iterator &it)
{
	it = _segments.begin();
	//cout << "Looking for the segment at the other side of vertex " << ver->getID() << " from segment " << seg->getID() << endl;
	for(int i=0; i<_nSegments; i++, it++)	{
		//cout << "Segment " << _segments[i]->getID() << " starts at " << _segments[i]->getStart()->getID() << " and ends at " << _segments[i]->getEnd()->getID() << endl;
		if(_segments[i]->getStart()->getID()==ver->getID() && _segments[i]->getID()!=seg->getID() )	{
			seg = _segments[i];
			ver = seg->getEnd();
			return;
		}
		else if(_segments[i]->getEnd()->getID()==ver->getID() && _segments[i]->getID()!=seg->getID())	{
			seg = _segments[i];
			ver = seg->getStart();
			return;
		}
	}
}

//This is called by splitRegion to remove the segments that has been moved to the new region

void Region::removeSegments(Vertex* start, Vertex* end)
{
	vector<Segment*>::iterator it = _segments.begin();
	int i = 0;
	Vertex* ver;
	Segment* seg;
	while(_segments[i]->getStart()->getID()!=start->getID() &&
			_segments[i]->getEnd()->getID()!=start->getID() &&
			i < _nSegments)	{
		i++;
		it++;
	}
	seg = _segments[i];
	//cout << "Removing segment " << seg->getID() << " from region " << _ID << endl;
	_segments.erase(it);
	ver = _segments[i]->getEnd();
	while(ver->getID()!=end->getID())	{
		getNextSegment(seg, ver, it);
		//cout << "Removing segment " << seg->getID() << " from region " << _ID << endl;
		_segments.erase(it);
	}
}

//Get the vertices in the order they surround the region
//Called by drawScaffold.

vector<Vertex*> Region::getOrderedVertices()
{
	Segment* seg = _segments[0];
	Vertex* ver = seg->getStart();
	vector<Vertex*> vers;
	vers.push_back(ver);
	for(int i=1; i<_nVertices; i++)	{
		getNextSegment(seg, ver);
		vers.push_back(ver);
	}
	return vers;
}

void Region::printInfo()
{
	cout << "Region " << _ID << " is surrounded by:" << endl << "Segments: ";
	for(int i=0; i<_segments.size(); i++)
		cout << _segments[i]->getID() << ", ";
	cout << endl << " Vertices: ";

	for(int j=0; j<_vertices.size(); j++)
		cout << _vertices[j]->getID() << ", ";
	cout << endl;
}

///////////////////////////////////////////////
//
//	Saving and loading
//
///////////////////////////////////////////////

void Region::saveRegion(ofstream &file)
{
	//file << _ID << endl;
	file << _segments.size() << endl;
	for(int i=0; i<_segments.size(); i++)
		file << _segments[i]->getID() << " ";
	file << endl;
	file << _vertices.size() << endl;
	for(int j=0; j<_vertices.size(); j++)
		file << " " << _vertices[j]->getID();
	file << endl;
}

void Region::loadRegion(ifstream &file, vector<Segment*> &segs, vector<Vertex*> &vers)
{
	cout << "Loading region " << _ID << endl;
	int ID, size;
	//char buf[80];
	//file.getline(buf, 80);
	//size = atoi(buf);
	file >> size;
	for(int i=0; i<size; i++)	{
		file >> ID;
		//cout << "Found seg with ID " << ID << endl;
		_segments.push_back(segs[ID]);
	}
	file >> size;
	for(int j=0; j<size; j++)	{
		file >> ID;
		//cout << "Found ver with ID " << ID << endl;
		_vertices.push_back(vers[ID]);
	}
}

