//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Scaffold.cpp: implementation of the Scaffold class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <sstream>
#include <fstream>
#include <maya/MGlobal.h>
#include "Scaffold.h"
#include "MELUtilities.h"
#include "math.h"
#include "stdlib.h"
#include <string>
#include "RewriteRule.h"
#include "Edge.h"
#include "Predecessor.h"
#include "Successor.h"
#include "World.h"
#include "BNF.h"
#include "Word.h"
#include "Axiom.h"
#include "Sentence.h"
#include "MELUtilities.h"
#include <maya/MFnMesh.h>
#include <maya/MFloatPointArray.h>
#include <maya/MIntArray.h>
#include <maya/MDoubleArray.h>
#include <maya/MFnNurbsSurface.h>
#include <maya/MPointArray.h>
#include <maya/MStatus.h>
#include <maya/MObject.h>
#include <maya/MString.h>
#include <maya/MFnDependencyNode.h>


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

//Use this when constructing the MEL-command for drawing lines

void streamPosition(Vector &pos, ostringstream &cmd)
{
	double tmp;
	cmd << "-p ";
	//cmd << "scmh -a ";
	tmp = pos.getX();
	if(tmp < 0)
		cmd << "-";
	cmd << fixed << fabs(tmp) << " ";
	tmp = pos.getY();
	if(tmp < 0)
		cmd << "-";
	cmd << fixed << fabs(tmp) << " ";
	tmp = pos.getZ();
	if(tmp < 0)
		cmd << "-";
	cmd << fixed << fabs(tmp) << " ";
}


Scaffold::Scaffold()
{
	_nCurves = 0;
	_nRegions = 0;
	_nSegments = 0;
	_nVertices = 0;
	_nSteps = 0;
	_nBranches = 0;
	_uKnots = 0;
	_vKnots = 0;
}

Scaffold::Scaffold(World* world, int ID, bool fixedPerimeter, bool fixedCenter, double pc, int pe, double limitedBranches, int branchAngle, bool softBoundary, int randRewrite, bool fullRandomRewrite, double branchLength, double angle)
{
	_ID = ID;
	_nCurves = 0;
	_nRegions = 0;
	_nSegments = 0;
	_nVertices = 0;
	_world = world;
	_fixedPerimeter = fixedPerimeter;
	_fixedCenter = fixedCenter;
	_pointConst = pc;
	_pointExponent = pe;
	_nSteps = 0;
	_nBranches = 0;
	_parsed = false;
	_syncronous = false;
	_angle = angle;
	_branchLength = branchLength;
	_branchAngle = branchAngle*PI/180;
	_softBoundary = softBoundary;
	_randRewrite = randRewrite;
	_nEdgeTypes = 1;
	_fullRandomRewrite = fullRandomRewrite;
	_uKnots = 0;
	_vKnots = 0;
}


Scaffold::~Scaffold()
{

}

void Scaffold::clear()
{
	//cout << "Clearing scaffold." << endl;
	_region.clear();
	_segments.clear();
	_vertices.clear();
	_curves.clear();
	_nRegions = 0;
	_nSegments = 0;
	_nVertices = 0;
	_nCurves = 0;
	_verticesInCurve.clear();
	_centerPos.setVector(0, 0, 0);
	_nSteps = 0;
	_nBranches = 0;
	_predecessorData.clear();
	_successorData.clear();
	_axiom->clear();
	_curveNames.clear();
}

//Calcualte the geometric center of the scaffold
//We use this when we grow so that each vertex can move away from the center

void Scaffold::geometricCenter()
{
	_maxX = -10000;
	_maxY = -10000;
	_maxZ = -10000;
	_minX = 10000;
	_minY = 10000;
	_minZ = 10000;

	for(int i=0; i<_nVertices; i++)	{
		_maxX = _maxX > _vertices[i]->getPosition().getX() ? _maxX : _vertices[i]->getPosition().getX();
		_minX = _minX < _vertices[i]->getPosition().getX() ? _minX : _vertices[i]->getPosition().getX();
		_maxY = _maxY > _vertices[i]->getPosition().getY() ? _maxY : _vertices[i]->getPosition().getY();
		_minY = _minY < _vertices[i]->getPosition().getY() ? _minY : _vertices[i]->getPosition().getY();
		_maxZ = _maxZ > _vertices[i]->getPosition().getZ() ? _maxZ : _vertices[i]->getPosition().getZ();
		_minZ = _minZ < _vertices[i]->getPosition().getZ() ? _minZ : _vertices[i]->getPosition().getZ();
	}
	_centerPos.setVector((_maxX+_minX)/2, (_maxY+_minY)/2, (_maxZ+_minZ)/2);
}

bool Scaffold::getCenter(Vector &pos)
{
	pos = _centerPos;
	return true;
}

/////////////////////////////////////////////
//
//	Constructors for the scaffold-elements
//
//	Use these constructors so that we get the correct IDs when constructing new elements
//
/////////////////////////////////////////////

//Use this to make a segment that has teh same parameter as oldSeg
//Called by splitsegment

void Scaffold::newSegment(Segment* &newSegment, Segment &oldSeg)
{
	Segment* newSeg = new Segment(_nSegments++, oldSeg);
	_segments.push_back(newSeg);
	newSegment = newSeg;
}

void Scaffold::newSegment(Segment* &newSegment)
{
	Segment* newSeg = new Segment(_nSegments++, true);
	_segments.push_back(newSeg);
	newSegment = newSeg;
}

//Called by joinBranches when we are going to subdivide a region

void Scaffold::newSegment(Segment* &newSegment, Vertex* start, Vertex* end, Region* reg)
{
	Segment* newSeg = new Segment(_nSegments++, false);
	newSeg->setEnd(end);
	newSeg->setStart(start);
	splitRegion(newSeg, reg, start, end);
	newSeg->setOrientation();
	_segments.push_back(newSeg);
	newSegment = newSeg;
}

void Scaffold::newVertex(Vertex* &newVertex, bool perimeter)
{
	Vertex* newVer = new Vertex(_nVertices++, _world->getBoundary(), _world->getForce(), _world->getNoise(), _branchLength, _branchAngle, false, perimeter);
	_vertices.push_back(newVer);
	newVertex = newVer;
}

void Scaffold::newVertex(Vertex* &newVertex, int curve, bool outside, bool perimeter)
{
	Vertex* newVer = new Vertex(_nVertices++, _world->getBoundary(), _world->getForce(), _world->getNoise(), _branchLength, _branchAngle, outside, perimeter);
	_vertices.push_back(newVer);
	//Update the number of vertices for this curve
	_verticesInCurve[curve]++;
	newVertex = newVer;
}

void Scaffold::newRegion(Region* &newRegion)
{
	Region* newReg = new Region(_nRegions++);
	_region.push_back(newReg);
	newRegion = newReg;
}

void Scaffold::newCurve(Segment* newCurve)
{
	newCurve->setCurve(_nCurves++);
	_curves.push_back(newCurve);
	_verticesInCurve.push_back(2);
}

//These are used when we load a scaffold from file
void Scaffold::createVertices()
{
	Vertex* ver;
	for(int i=0; i<_nVertices; i++)	{
		ver = new Vertex(i, _world->getBoundary(), _world->getForce(), _world->getNoise(), _branchLength, _branchAngle, false, false);
		_vertices.push_back(ver);
	}
}

void Scaffold::createSegments()
{
	Segment* seg;
	for(int i=0; i<_nSegments; i++)	{
		seg = new Segment(i, false);
		_segments.push_back(seg);
	}
}

void Scaffold::createRegions()
{
	Region* reg;
	for(int i=0; i<_nRegions; i++)	{
		reg = new Region(i);
		_region.push_back(reg);
	}
}

//Use this to create a seed from a curve that the user has provided in Maya

void Scaffold::setSeed(vector<Vector> cvs, vector<double> masses, int nEdgeTypes)
{
	Vertex *startVer, *prevVer, *newVer;
	Segment *prevSeg, *newSeg;
	Region *reg;
	Vector dir, norm;
	double length = 0;

	//Create a region, a vertex and a segment to begin with
	newRegion(reg);
	newVertex(startVer, true);
	prevVer = startVer;
	prevVer->setPosition(cvs[0]);
	//Currently, this is the only way to edit the mass of a vertex
	//It is done through Maya (points in Maya can have a mass)
	//and we simply get that attribute value.
	prevVer->setMass(masses[0]);
	newSegment(newSeg);
	newSeg->setStart(prevVer);
	//Unfortunately, you can not control how the edge types are assigned, it is done randomly
	newSeg->setType( rand() % nEdgeTypes );
	newCurve(newSeg);
	prevSeg = newSeg;
	reg->addSegment(newSeg);
	//Create the rest of the vertices and segments
	for(int i=1; i<cvs.size(); i++)	{
		//Calculate the characteristic length	
		length += (cvs[i-1]-cvs[i]).getLength();
		//Make sure that all the vertices and segments are connected properly
		prevVer->addSegment(newSeg);
		//Create a new segment based on the previous
		newSegment(newSeg, (*prevSeg));
		newVertex(newVer,0, false, true);
		reg->addSegment(newSeg);

		newSeg->setStart(newVer);
		newVer->setPosition(cvs[i]);
		newVer->setMass(masses[i]);
		
		dir = newVer->getPosition() - prevVer->getPosition();
		dir.normalize();
		prevSeg->setDirection(dir);
		newSeg->setType( rand() % nEdgeTypes );
		prevSeg->setEnd(newVer);
		newVer->addSegment(prevSeg);
		
		prevVer = newVer;
		prevSeg = newSeg;
	}
	//Make sure that the ends meet
	prevSeg->setEnd(startVer);
	prevVer->addSegment(newSeg);
	startVer->addSegment(newSeg);
	reg->addVertex(newVer);
	//Find out if the region is to the left or the right by taking a cross-product
	//We assume that the normal has a z-component!=0
	norm = _segments[0]->getDirection()&_segments[1]->getDirection();
	if(norm.getZ() < 0)	{
		for(int j=0; j<_nSegments; j++)	{ //At the start there are as many segments as vertices (dangerous)
			_segments[j]->setRight(reg);
			_vertices[j]->addRegion(reg);
		}
	}
	else	{
		for(int j=0; j<_nSegments; j++)	{
			_segments[j]->setLeft(reg);
			_vertices[j]->addRegion(reg);
		}
	}
	//Fix the center if necessary
	if(_fixedCenter)	{
		geometricCenter();
		_fixedCenterPos = _centerPos;
	}
	_characteristicLength /= cvs.size();
	//printVertices();
	//printSegments();
}

//Use this function for the genotypically encoded axioms
void Scaffold::setSeed(Vector startPos, double length)
{
	//cout << "Setting seed, length is " << length << endl;
	//Parse the axiom, we can not do it earlier because we do not know if the user will provide us with a seed
	int nSegs = _axiom->numberOfWords(EDGE);
	//Do not use the angle defined in the genome
	double angle = 2*PI/nSegs;
	vector<bool> changeDirs;
	bool changeDir = false, turn = false;
	Vertex *startVer, *prevVer, *newVer;
	Segment *prevSeg, *newSeg;
	Region *reg;
	//We always draw the first edge in this direction
	Vector dir(START_DIR_X, START_DIR_Y, START_DIR_Z);
	Word *word;
	//Create the first region, segment and vertex
	word = _axiom->getWord(0);
	newRegion(reg);
	newVertex(startVer, true);
	prevVer = startVer;
	newSegment(newSeg);
	newSeg->setStart(prevVer);
	newSeg->setDirection(dir);
	newSeg->setLength(length);
	newSeg->setRight(reg);
	newSeg->setType( word->getEdgeType() );
	newCurve(newSeg);
	reg->addSegment(newSeg);
	prevSeg = newSeg;
	changeDirs.push_back(false);
	//Parse the axiom and add the rest of 
	for(int i=1; i<_axiom->numberOfWords(); i++)	{
		word = _axiom->getWord(i);
		//cout << "Found word " << word->getName() << endl;
		if(word->getName()=="+")
			turn = true;
		else if(word->getName()=="~")
			changeDir = !changeDir;
		else if(word->getName().find(EDGE)!=-1)	{
			prevVer->addSegment(newSeg);

			newSegment(newSeg, (*prevSeg));
			newVertex(newVer,0 , false, true);
			reg->addSegment(newSeg);

			newSeg->setStart(newVer);
			newVer->setPosition(prevVer->getPosition() + dir*length);
			if(turn)
				dir.turn(angle);
			
			turn = false;
			newSeg->setDirection(dir);
			newSeg->setType( word->getEdgeType() );
			newCurve(newSeg);
			prevSeg->setEnd(newVer);
			newVer->addSegment(prevSeg);
			
			prevVer = newVer;
			prevSeg = newSeg;
			changeDirs.push_back(changeDir);
			changeDir = false;
		}
	}
	//Make sure that it all fits in the end
	prevSeg->setEnd(startVer);
	prevVer->addSegment(newSeg);
	startVer->addSegment(newSeg);
	reg->addVertex(newVer);
	_characteristicLength = length;
	//Change directions (not a very nice solution...)
	//cout << "Changing directions, there are " << _nSegments << " segments and changeDirs has length " << changeDirs.size() << endl;
	for(int j=0; j<nSegs; j++)	{
		if(changeDirs[j])	
			_segments[j]->switchDirection();
	}
	//Move to the desired startpos
	moveCenter(startPos);
	if(_fixedCenter)
		_fixedCenterPos = startPos;
	//These are only used for the predefined 4-sided grammar
	_uKnots = 2;
	_vKnots = 2;
	//printVertices();
	//printSegments();
}

////////////////////////////////////////////////
//
//	Printing
//
////////////////////////////////////////////////

void Scaffold::printSegments()
{
	for(int l=0; l<_segments.size(); l++)	{
		cout << "Segment " << _segments[l]->getID() << " starts at " << _segments[l]->getStart()->getID() << " and ends at " << _segments[l]->getEnd()->getID();
		cout << ", it has type " << _segments[l]->getType();
		cout << " and has length " << _segments[l]->getLength() << endl;
		cout << "The direction is " << _segments[l]->getDirection().printVector().c_str() << endl;
	}
}

void Scaffold::printVertices()
{
	for(int j=0; j<_nVertices; j++)	{
		cout << "Vertex " << _vertices[j]->getID() << " is located in " << _vertices[j]->getPosition().printVector().c_str() << " and has segments: ";
		for(int k=0; k<_vertices[j]->numberOfSegments(); k++)
			cout << _vertices[j]->getSegment(k)->getID() << " ";
		cout << endl << "It is surrounded by the following regions: ";
		for(int i=0; i<_vertices[j]->numberOfRegions(); i++)
			cout << _vertices[j]->getRegion(i)->getID() << " ";
		cout << endl;
	}
}

void Scaffold::printCurve(int i)
{
	Vertex *currentVer = _curves[i]->getStart();
	Segment *currentSeg = _curves[i];
	
	cout << "Curve " << i << " consists of:" << endl;
	//cout << "Vertex " << currentVer->getID() << endl;
	for(int j=0; j<_verticesInCurve[i]; j++)	{
		cout << "Vertex " << currentVer->getID() << endl;
		nextVertex(i, currentVer, currentSeg);
	}	
}

void Scaffold::printRegions()
{
	for(int k=0; k<_nRegions; k++)
		_region[k]->printInfo();
}

//This function takes the scaffold and turns it into a Maya-object.
//It will either create a bunch of curves, or a mesh-surface

void Scaffold::drawScaffold(surfaceRepresentation surfRep)
{
	if(surfRep==QUILT && _nVertices==_uKnots*_vKnots)	{	//Make sure that we have all the CVs, otherwise it is not possible to create the surface
		//This is only used for the predefined grammar
		//Create the array of cvs
		vector<Vertex*> cvArray = updateCVArray();
		MPointArray cvs;
		for(int i=0; i<_nVertices; i++)
			cvs.append(cvArray[i]->getPosition().getX(), cvArray[i]->getPosition().getY(), cvArray[i]->getPosition().getZ());
		//Create the knot vectors
		//If we have enough vertices, we can make it degree so that it look nicer
		MDoubleArray uknots, vknots;
		for(int j=0; j<(_nSteps > 2 ? _uKnots + 2: _uKnots); j++)
			uknots.append(j);
		for(int k=0; k<(_nSteps > 2 ? _vKnots + 2: _vKnots); k++)
			vknots.append(k);
		MFnNurbsSurface surf;
		MStatus *status = new MStatus();
		//Create the actual surface object, it is a real nurbs-surface
		MObject object;
		if(_nSteps>2)
			object = surf.create(cvs, uknots, vknots, 3, 3, MFnNurbsSurface::kOpen, MFnNurbsSurface::kOpen, false, MObject::kNullObj, status);
		else
			object = surf.create(cvs, uknots, vknots, 1, 1, MFnNurbsSurface::kOpen, MFnNurbsSurface::kOpen, false, MObject::kNullObj, status);
		if(*status==MStatus::kFailure)	//We might fail anyway
			cout << "Sorry, failed to produce a real NURBS-surface." << endl << "Try again with an increased the BranchAngle." << endl << "If that does not work" << endl << ", you will probably have to change the environment to make the surface smoother." << endl;
		else
			MString str = ((MFnDependencyNode)object).setName("genr8Surface#", status);
	}
	else if(surfRep==SURFACE || surfRep==QUILT)	{
		if(surfRep==QUILT)
			cout << "Sorry, failed to produce a real NURBS-surface." << endl << "Try again with an increased the BranchAngle." << endl << "If that does not work" << endl << ", you will probably have to change the environment to make the surface smoother." << endl;
		//Use Maya's built in algorithm for creating a mesh from a point cloud. The only problem is that it isn't a nurbs-surface.
		//Vertices
		MFloatPointArray vertexArray;
		for(int i=0; i<_nVertices; i++)
			vertexArray.append(_vertices[i]->getPosition().getX(), _vertices[i]->getPosition().getY(), _vertices[i]->getPosition().getZ());
		//Number of vertices for each polygon
		//How the regions are connected to the vertices
		MIntArray polygonCount, polygonConnect;
		int nVer;
		vector<Vertex*> vers;
		for(int j=0; j<_nRegions; j++)	{
			//Get the number of vertices for the region and store it
			nVer = _region[j]->numberOfVertices();
			vers = _region[j]->getOrderedVertices();
			polygonCount.append(nVer);
			for(int k=0; k<nVer; k++)
				polygonConnect.append(vers[k]->getID());
		}
		//Create the surface
		MFnMesh mesh;
		MObject object;
		MStatus *status = new MStatus();
		string* name = new string;
		object = mesh.create(_nVertices, _nRegions, vertexArray, polygonCount, polygonConnect);
		MString str = ((MFnDependencyNode)object).setName("genr8Surface#", status);
		name->assign(str.asChar());
		_curveNames.push_back( name );
	}
	else	{
		//Draw lines to create a scaffold model.
		//This is by far the easiest and we can represent arbitrary surfaces
		//Unfortunately, it is not practical for the user and it also takes a lot of time compared to the surface
		Vertex *currentVer;
		Segment *currentSeg;
		string* curveName;
		
		//Each curve is a line
		for(int i=0; i<_nCurves; i++)	{
			ostringstream splinecmd, cmd;
			currentVer = _curves[i]->getStart();
			currentSeg = _curves[i];
			
			if(surfRep==BSPLINES)
				cmd << "curve -d 1 -n tmpCurve";
			else
				cmd << "curve -d 1 -n genr8Curve" << i << " ";
			
			streamPosition(currentVer->getPosition(), cmd);
			currentVer = currentSeg->getEnd();
			for(int j=1; j<_verticesInCurve[i]; j++)	{
				//Get the position
				streamPosition(currentVer->getPosition(), cmd);
				//Go to next vertex
				nextVertex(i, currentVer, currentSeg);
			}

			//Make it smooth
			if(surfRep==BSPLINES)	{
				eval(cmd.str());
				splinecmd << "fitBspline -ch 0 -tol 0.001 -n genr8Curve" << i << " ";
				curveName = getStringPointerResults(splinecmd.str());
				eval("delete tmpCurve");
			}
			else
				curveName = getStringPointerResult(cmd.str());
			//Store the name			
			_curveNames.push_back( curveName );
		}
	}
	//printRegions();
	//printSegments();
	//printVertices();
}

//This function makes the region more uniform by moving the vertices so that they try to maximize
//the distance to the neighbours. This can have a serious effect on the resulting subdivisions.

bool Scaffold::uniformRegions(double minDifference, double relMinDifference)	//The second parameter is not used
{
	double minDiff = 0, disp;
	int j = 0;
	bool closeEnough = false;
	//Move the veertices until they are close enogh to the optimal point or until we feel that it is not worth doing it anymore
	while(!closeEnough && j < 10)	{
		//cout << "Trying to make regions uniform, minDiff is " << minDiff << endl;
		minDiff = 0;
		for(int i=0; i<_nVertices; i++)	{
			if(!_vertices[i]->onPerimeter())	{	//Fix the ones on the perimeter
				disp = _vertices[i]->moveToCenterOfNeighbours();
				minDiff =  disp > minDiff ? disp : minDiff;
			}
		}
		closeEnough = minDiff < minDifference;
		j++;
	}
	//cout << "Regions uniform after " << j << " iterations, minDiff is " << minDiff << endl;
	return true;
}

//Use this to move along a specific curve
//Starting from vertex &current which will contain the next vertex when the function returns
//&segment contains the segment we are traversing along and when the function returns,
//it will contain the segment `on the other side`.


bool Scaffold::nextVertex(int curve, Vertex* &current, Segment* &segment)
{
	Vertex *ver;
	//Move to the end of the segment
	if(!current->getNextOnCurve(curve, ver, segment))
		return false;
	
	current = ver;
	return true;
}

//This is used when we are trying to construct the CV-matrix for the nurbs-surface.
//Instead of following a curve, we follow odd or even edge types.
//This is possible because we know the topology of the surface

bool Scaffold::nextVertex(Vertex* &current, Segment* &segment, bool oddType)
{
	Segment* seg;
	//cout << "Looking for the next vertex after " << current->getID() << endl;
	for(int i=0; i<current->numberOfSegments(); i++)	{
		seg = current->getSegment(i);
		//cout << "Found segment " << seg->getID() << " with type " << seg->getType() << endl;
		//When we start, we do not have a segment yet
		if(!segment)	{
			if((seg->getType() % 2)==(int)oddType)	{
				segment = seg;
				if(seg->getEnd()->getID()==current->getID())
					current = seg->getStart();
				else
					current = seg->getEnd();
				return true;
			}
		}
		else	{
			if((seg->getType() % 2)==(int)oddType && seg->getID()!=segment->getID())	{
				segment = seg;
				if(seg->getEnd()->getID()==current->getID())
					current = seg->getStart();
				else
					current = seg->getEnd();
				return true;
			}
		}
	}
	return false;
}

//This function loops through all the branches and tries to match them

void Scaffold::joinBranches()
{
	bool startVer;
	Region* reg;
	Vertex *start, *end, *tmpVer;
	Segment* newSeg;
	vector<Region*> regions;
	int type;
	
	//There are four nested loops here, but I believe that it is more efficient than a simple n^2 algorithm that compares
	//all the vertices.
	//cout << "There are " << _nBranches << " branches." << endl;
	if(_nBranches>1)	{
		//The branches are owned by the vertices, so this allows us to go through all the branches
		for(int i=0; i<_nVertices; i++)	{
			//cout << "Resulting force before matchBranch: " << _vertices[i]->_resultingForce.printVector().c_str() << endl;
			//cout << "Getting adjacent regions for vertex " << _vertices[i]->getID() << " with " << _vertices[i]->numberOfRegions() << " regions." << endl;
			//_vertices[i]->printInfo();
			_vertices[i]->adjacentRegions(regions);
			//We must check all the adjacent regions for neighbours
			for(int j=0; j<regions.size(); j++)	{
				reg = _vertices[i]->getRegion(j);
				//cout << "Got region " << reg->getID() << " with " << reg->numberOfVertices() << " vertices." << endl;
				//Go through the vertices of each region
				for(int k=0; k<regions[j]->numberOfVertices(); k++)	{
					tmpVer = reg->getVertex(k);
					//start = reg->getVertex(k);
					//cout << "Got vertex " << tmpVer->getID() << " with " << tmpVer->numberOfBranches() << " branches." << endl;
					if(tmpVer->getID()!=_vertices[i]->getID())	{
						//Go through all the branches on those vertices
						for(int l=0; l<tmpVer->numberOfBranches(); l++)	{
							//Check for match
							if(_vertices[i]->matchBranch(tmpVer->getBranch(l), type, reg, startVer, _centerPos, _characteristicLength))	{
								//Update the scaffold immediately, it is not possible to do all the matchings first and then connect the branches
								start = _vertices[i];
								end = tmpVer;
								//cout << "Joined branches during step " << _nSteps << endl;
								//cout << "Joining branches in region " << reg->getID();
								//cout << " between vertex " << start->getID() << " and " << end->getID() << endl;
								newSegment(newSeg, start, end, reg);
								newSeg->setType(type);
								newCurve(newSeg);
								start->addSegment(newSeg);
								end->addSegment(newSeg);
							}
						}
					}
				}
			}
		}
	}
	//cout << "Finished joining branches." << endl;
}

//This function is called by newSegment to split an old region

void Scaffold::splitRegion(Segment* &divider, Region* oldReg, Vertex* start, Vertex* end)
{
	//cout << "Splitting region " << oldReg->getID() << " with segment " << divider->getID() << endl;
	vector<Segment*> segs;
	vector<Vertex*> vers;
	Region *newReg, *right, *left;
	//Find which vertices should form the border
	oldReg->splitRegion(start, end, segs);
	//Create new region
	newRegion(newReg);
	//Both regions should have the new segment
	newReg->addSegment(divider);
	oldReg->addSegment(divider);
	//Find out if the old region is to the left or to the right of the new segment
	//The first segment is connected to start
	right = segs[0]->getRight();
	if(segs[0]->getEnd()->getID()==start->getID())	{
		if(right!=NULL)	{
			if(segs[0]->getRight()->getID()==oldReg->getID())	{
				divider->setLeft(oldReg);
				divider->setRight(newReg);
			}
			else	{
				divider->setRight(oldReg);
				divider->setLeft(newReg);
			}
		}
		else	{
			if(segs[0]->getLeft()->getID()==oldReg->getID())	{
				divider->setRight(oldReg);
				divider->setLeft(newReg);
			}
			else	{
				divider->setLeft(oldReg);
				divider->setRight(newReg);
			}
		}
	}
	else	{
		if(right!=NULL)	{
			if(right->getID()==oldReg->getID())	{
				divider->setRight(oldReg);
				divider->setLeft(newReg);
			}
			else	{
				divider->setLeft(oldReg);
				divider->setRight(newReg);
			}
		}
		else	{
			if(segs[0]->getLeft()->getID()==oldReg->getID())	{
				divider->setLeft(oldReg);
				divider->setRight(newReg);
			}
			else	{
				divider->setRight(oldReg);
				divider->setLeft(newReg);
			}
		}
	}
	//cout << "Setting region " << divider->getLeft()->getID() << " to the left and region " << divider->getRight()->getID() << " to the right." << endl;
	//Add those segments to the new region and add the region to the correct side of the segments
	for(int i=0; i<segs.size(); i++)	{
		newReg->addSegment(segs[i]);
		segs[i]->getEnd()->addRegion(newReg);
		segs[i]->getStart()->addRegion(newReg);
		right = segs[i]->getRight();
		left = segs[i]->getLeft();
		if(right!=NULL)	{
			if(right->getID()==oldReg->getID())
				segs[i]->setRight(newReg);
			else
				segs[i]->setLeft(newReg);
		}
		if(left!=NULL)	{
			if(left->getID()==oldReg->getID())
				segs[i]->setLeft(newReg);
			else
				segs[i]->setRight(newReg);
		}
	}
	//newReg->printInfo();
	//oldReg->printInfo();
}

//Use this function to erase the Maya drawings

void Scaffold::eraseDrawings()
{
	ostringstream cmd;
	//The curve names were stored when they were created
	for(int i=0; i<_curveNames.size(); i++)	{
		//cout << "Deleting curve " << _curveNames[i]->c_str() << endl;
		cmd << "delete " << _curveNames[i]->c_str() << ";";
	}
	
	eval(cmd.str());
	_curveNames.clear();
}

//Use this function at the end of each step do do some cleanup and bookkeeping

void Scaffold::stepFinished(int maxScale)
{
	//Clear some stuff, remove all the branches that were not matched and restore the flags of the segments
	_nBranches = 0;
	for(int i=0; i<_nSegments; i++)
		_segments[i]->setRewritten(false);
	for(int j=0; j<_nVertices; j++)
		_vertices[j]->removeAllBranches();
	//Increase some counters
	_nSteps++;
	_characteristicLength *= maxScale;
	_uKnots = 2*_uKnots - 1;
	_vKnots = 2*_vKnots - 1;	
}

//Use this to move the entire scaffold

void Scaffold::moveCenter(Vector pos)
{
	Vector tmp, diff;
	geometricCenter();
	diff = pos - _centerPos;
	_centerPos = pos;
	for(int i=0; i<_vertices.size(); i++)	{
		tmp = _vertices[i]->getPosition();
		_vertices[i]->setPosition(tmp + diff);
	}
}

void Scaffold::repellPoints()
{
	double massratio;
	Vector disp, zero(0, 0, 0);
	vector<Vector> disps;
	disps.resize(_nVertices);
	//Expensive calculation
	for(int i=0; i<_nVertices; i++)	{
		for(int j=i+1; j<_nVertices; j++)	{
			disp = _vertices[i]->getForce(_vertices[j]->getPosition(), _pointConst, _pointExponent);
			//cout << "The force between " << i << " and " << j << " is " << disp.printVector().c_str() << endl;
			massratio = _vertices[j]->getMass() / _vertices[i]->getMass();
			disps[i] += disp*massratio;
			disps[j] += massratio!=0 ? disp*(-1/massratio) : zero;
		}
	}
	//Move the points
	for(int k=0; k<_nVertices; k++)	{
		if(!(_fixedPerimeter && _vertices[k]->onPerimeter()))	//If we have fixed the perimeter, we should not move those vertices
			_vertices[k]->moveVertex(disps[k], true);
	}
}

/////////////////////////////////////////////
//
//	Growth
//
///////////////////////////////////////////////

//This function is called when we want to grow one step
//We are allowed to grow by different amounts in the different directions

void Scaffold::step(double scaleX, double scaleY, double scaleZ)
{
	vector<Segment*>	preds;
	growScaffold(scaleX, scaleY, scaleZ);
	vector<SuccessorData> succDatas;
	double maxScale = scaleX < scaleY ? scaleY : scaleX;
	maxScale = maxScale < scaleZ ? scaleZ : maxScale;
	//cout << "PredecessorData has length " << _predecessorData.size() << " and SuccessorData has length " << _successorData.size() << endl;
	for(int i=0; i<_predecessorData.size(); i++)	{
		//Get the segments that we shall rewrite
		segmentsToRewrite(preds, _predecessorData[i]);
		//Rewrite them
		getSuccessors(i, succDatas);
		rewrite(preds, succDatas);
		
		if(!_syncronous)	{
			//Match the branches
			joinBranches();
		}
	}
	if(_syncronous)	{
		//Match the branches
		joinBranches();
	}
	stepFinished(maxScale);
}

//Get the successor for this production, if we have stochastic productions, there are several to choose from
//predecessor is the index to the production we are currently applying.

void Scaffold::getSuccessors(int predecessor, vector<SuccessorData> &succDatas)
{
	int succ, start = 0;
	succDatas.clear();
	for(int i=0; i<predecessor; i++)
		start += _predecessorData[i]._nSucc;
	//Get the successors that corresponds to this predecessor
	if(_fullRandomRewrite)	{
		for(int j=start; j<start + _predecessorData[predecessor]._nSucc; j++)
			succDatas.push_back(_successorData[j]);
	}
	else	{
		//Just return one rewrite rule
		succ = _predecessorData[predecessor]._nSucc==1 ? predecessor : start + rand() % _predecessorData[predecessor]._nSucc;
		succDatas.push_back( _successorData[succ] );
	}
	//cout << "There are " << _predecessorData[predecessor]._nSucc << " possible successors, returning " << succ - start << endl;
}

//This function moves the vertices so that the scaffold grows

void Scaffold::growScaffold(double scaleX, double scaleY, double scaleZ)
{
	//Calcualte the location of the center
	geometricCenter();
	//Scale up
	for(int i=0; i<_nVertices; i++)	{
		if(!(_vertices[i]->onPerimeter() && _fixedPerimeter))
			_vertices[i]->scale(scaleX, scaleY, scaleZ, _centerPos, _softBoundary);
	}
	//Do the repelling afterwards, perhaps not optimal...
	if(_pointConst!=0)
		repellPoints();
	//Calcualte the new location of the center
	geometricCenter();
	//IF iit is fixed, move it back
	if(_fixedCenter)
		moveCenter(_fixedCenterPos);
	//Recalculate the directions and lengths of the segments
	for(int j=0; j<_nSegments; j++)
		_segments[j]->setOrientation();
}

//This function searches the list of segments and puts the ones that should be rewritten in the &segs array.

void Scaffold::segmentsToRewrite(vector<Segment*> &segs, PredecessorData predData)
{
	segs.clear();
	Segment* seg;
	//cout << "Looking for segments to rewrite with type " << predData._type << " preceded by type " << predData._prevType << " and followed by type " << predData._follType << endl;
	//Loop through all segments
	for(int i=0; i<_nSegments; i++)	{
		seg = _segments[i];
		//Find segments of correct type
		if( seg->getType()==predData._rewrite._type && !seg->isRewritten() )	{
			if(seg->checkCounter(predData._rewrite))	{
			//if( checkCounter(seg->getCounter(), predData._rewrite) )	{
				//Check to see if the context is correct
				if(predData._prev._type!=-1)	{
					if(!seg->isTypeAndCountPrevious(predData._prev))
						continue;
				}
				if(predData._foll._type!=-1)	{
					if(!seg->isTypeAndCountFollowing(predData._foll))
						continue;
				}
				//The segment should be rewritten, add it to the array
				segs.push_back(seg);
			}
		}
	}
	//cout << "Returning " << segs.size() << " segments." << endl;
}

//This function applies a production to the segments in &preds
//and it uses the rules contained in succDatas

void Scaffold::rewrite(vector<Segment*> &preds, vector<SuccessorData> succDatas)
{
	vector<Segment*> splits;
	int nPreds = preds.size(), parts, nSucc = 1;
	Vector startDir, orgDir;
	double newLength, oldLength;
	SuccessorData succData;
	//Do we have fullRandomRewrite, ie a random production for each segment
	if(succDatas.size()==1)
		succData = succDatas[0];
	else
		nSucc = succDatas.size();
	//Apply production to each segment
	for(int i=0; i<nPreds; i++)	{
		//Split the segments
		oldLength = preds[i]->getLength();
		if(nSucc!=1)
			//Choose a rule at random
			succData = getRandomSuccessor(succDatas);

		//Check the condition
		//if(!checkCondition(preds[i]->getCounter(), succData._conditionData))
		if(!preds[i]->checkCondition(succData._conditionData))
			continue;
		parts = succData._edges.size();
		//Split the segment into the specified number of parts
		splitSegments(splits, preds[i], parts);
		startDir = splits[0]->getDirection();
		orgDir = startDir;
		newLength = 0;
		//Transform them and place branches
		//cout << "Segment " << preds[i]->getID() << " was split into " << parts << " parts." << endl;
		for(int j=0; j<parts; j++)	{
			//Change type
			//cout << "Setting type of segment " << splits[j]->getID() << " to " << succData._edges[j]._type << endl;
			splits[j]->setType( rand() % 100 < _randRewrite ? rand() % _nEdgeTypes : succData._edges[j]._type._type);
			//Change counter
			if(succData._edges[j]._type._counterType==I_PLUS_ONE)
				splits[j]->incCounter();
			else if(succData._edges[j]._type._counterType==I_MINUS_ONE)
				splits[j]->decCounter();
			else if(succData._edges[j]._type._counterType==NUMERIC)
				splits[j]->setCounter(succData._edges[j]._type._counter);
			//Change orientation
			splits[j]->changeOrientation(succData._edges[j]._operators, _angle, startDir);
			newLength += orgDir*startDir;
			//Place branches
			//cout << "Placing branches on vertex " << splits[j]->getID() << ", there are " << succData._edges[j]._branches.size() << " breanches" << endl;
			for(int l=0; l<succData._edges[j]._branches.size(); l++)
				splits[j]->getStart()->placeBranch(succData._edges[j]._branches[l], startDir, _angle, _nBranches++);
			//Mark the segment, if we are using asyncronous mode, we may rewrite it again during this step
			if(_syncronous)
				splits[j]->setRewritten(true);
		}
		//Adjust length
		newLength = oldLength/newLength;
		splits[0]->changeLength(newLength, parts);
	}
}

//Called by rewrite. It will give a random successor for this predecessor. Used only for fullRandomRewrite

SuccessorData Scaffold::getRandomSuccessor(vector<SuccessorData> succDatas)
{
	//Calculate the total weight
	int nSucc = succDatas.size(), totalWeight = 0, succ, j = 0, weight = 0;
	for(int i=0; i<nSucc; i++)
		totalWeight += succDatas[i]._weight;
	//Get a random weighted successor
	succ = rand() % totalWeight;
	while(weight<succ && j<nSucc)
		weight += succDatas[j++]._weight;

	return succDatas[j];
}

//Splits pred into parts new segments that are returned in &splits
//This is used to genreate the new segments that are the result of the prodution

void Scaffold::splitSegments(vector<Segment*> &splits, Segment* pred, int parts)
{
	Segment *newSeg, *prevSeg, *originalSeg;
	Vertex *newVer, *originalVer, *endVer;
	Region *left, *right;
	
	double step = pred->getLength()/parts;
	Vector dir;
	splits.clear();
	//Update the segments and the vertices about the new state of the scaffold and make sure that they have the correct info.
	if(parts > 1)	{
		//cout << "Splitting segment " << pred->getID() << " of type " << pred->getType() << " into " << parts << " parts" << endl;
		//cout << "The original segment started in " << pred->getStart()->getPosition().printVector().c_str() << endl << " and ends in " << pred->getEnd()->getPosition().printVector().c_str() << endl;
		left = pred->getLeft();
		right = pred->getRight();

		endVer = pred->getEnd();
		endVer->removeSegment(pred);

		pred->setLength(step);
		
		originalSeg = pred;
		originalVer = originalSeg->getStart();
		dir = originalSeg->getDirection();
		dir.normalize();
		splits.push_back(originalSeg);
		prevSeg = originalSeg;
		
		//Create new segments and vertices
		for(int i=1; i < parts; i++)	{
			newSegment(newSeg, (*originalSeg));
			newVertex(newVer, originalSeg->getCurve(), endVer->isOutside(), originalSeg->isPerimeter());
			
			newSeg->setStart(newVer);
			if(left!=NULL)
				left->addSegment(newSeg);
			if(right!=NULL)
				right->addSegment(newSeg);
			
			newVer->setPosition(originalVer->getPosition() + dir*(step*i));
			//cout << "Vertex created in " << newVer->getPosition().printVector().c_str() << endl;
			newVer->addSegment(newSeg);
			newVer->addSegment(prevSeg);
			
			prevSeg->setEnd(newVer);
			splits.push_back(newSeg);
			prevSeg = newSeg;
		}
		//Restore the end
		newSeg->setEnd(endVer);
		endVer->addSegment(newSeg);
		
	}
	else
		splits.push_back(pred);
	//cout << "Finished splitting, returning " << splits.size() << " segments." << endl;
}

//This function is used to create the array of CVs that is required by Maya to create a nurbs-surface
//The strategy is to traverse along one of the edges of the perimeter (the for-loop) and for each vertex go down to the other side (the while-loop).

vector<Vertex*> Scaffold::updateCVArray()
{
	vector<Vertex*> cvArray;
	Vertex *currentVer, *curveVer = _curves[0]->getStart();
	Segment *currentSeg = NULL, *curveSeg = _curves[0];
	bool oddType = (curveSeg->getType() % 2)==0;
	
	currentVer = curveVer;
	cvArray.push_back(currentVer);
	while(nextVertex(currentVer, currentSeg, oddType))
		//Store in the matrix
		cvArray.push_back(currentVer);
	curveVer = curveSeg->getEnd();
	//Loop in one direction
	for(int i=1; i<_verticesInCurve[0]; i++)	{
		currentVer = curveVer;
		cvArray.push_back(currentVer);
		//Loop in the other direction
		while(nextVertex(currentVer, currentSeg, oddType))
			//Store in the matrix
			cvArray.push_back(currentVer);
		//Go to the next vertex in the other direction
		nextVertex(0, curveVer, curveSeg);
	}
	return cvArray;
}

///////////////////////////////////////////////////////////
//
//Parsing
//
//	Before we can start growing the scaffold, we must parse the sentence to extract the necessary
//	information in a more suitable format than that of the grammar.
//
///////////////////////////////////////////////////////////

//This function is called by the individual when it is time to parse the grammar

void Scaffold::parseSentence(Sentence* sentence)
{
	//The first line is always the axiom
	int startPos = 0, endPos = sentence->findWord(END_VAR, 0);
	Sentence *sent = sentence->getSentence(startPos, endPos);
	//We must save the axiom and parse it later...
	_axiom = sent;
	//setSeed(sent);
	startPos = endPos + 1;
	endPos = sentence->findWord(END_VAR, startPos);
	//Get the rewrite rules and the parameters
	while(endPos!=-1)	{
		sent = sentence->getSentence(startPos, endPos);
		if(sent->getWord(0)->getName()=="Angle")
			parseAngle(sent);
		else if(sent->getWord(0)->getName()=="Sync")
			setSyncronous(true);
		else if(sent->getWord(0)->getName()=="BranchAngle")
			setBranchAngle( atoi(sent->getWord(1)->getName().c_str()) );
		else if(sent->isRewriteRule())
			parseRewriteRule(sent);

		startPos = endPos + 1;
		endPos = sentence->findWord(END_VAR, startPos);
	}
	_parsed = true;
}

//We know the structures of successors; use the info when we parse it.

SuccessorData Scaffold::parseSuccessor(Sentence* succ)
{
	BranchData branchData;
	branchData.clear();
	SuccessorData succData;
	succData.clear();
	EdgeData edgeData;
	edgeData.clear();
	EdgeTypeAndCount edge;
	edge.clear();
	string name;
	int nWords = succ->numberOfWords(), stackDepth = 0, weightFound = -1, conditionFound = -1;
	//cout << "Parsing successor." << endl;
	for(int i=0; i<nWords; i++)	{
		name = succ->getWord(i)->getName();
		//cout << "Found term " << name.c_str() << endl;
		if(name.substr(0,4)==EDGE)	{
			if(stackDepth==0)	{	//Split
				succ->getWord(i)->getEdgeTypeAndCount(edge);
				edgeData._type = edge;
				succData._edges.push_back(edgeData);
				edgeData.clear();
			}
			else	{	//Branch
				branchData._type = succ->getWord(i)->getEdgeType();
				edgeData._branches.push_back(branchData);
				branchData.clear();
			}
		}
		else if(name=="[")	//The comming items should be pushed on the stack
			stackDepth++;
		else if(name=="]")
			stackDepth--;
		else if(name=="+")	{
			if(stackDepth==0)
				edgeData._operators.push_back(TURN_RIGHT);
			else
				branchData._operators.push_back(TURN_RIGHT);
		}
		else if(name=="-")	{
			if(stackDepth==0)
				edgeData._operators.push_back(TURN_LEFT);
			else
				branchData._operators.push_back(TURN_LEFT);
		}
		else if(name=="&")	{
			if(stackDepth==0)
				edgeData._operators.push_back(PITCH_DOWN);
			else
				branchData._operators.push_back(PITCH_DOWN);
		}
		else if(name=="^")	{
			if(stackDepth==0)
				edgeData._operators.push_back(PITCH_UP);
			else
				branchData._operators.push_back(PITCH_UP);
		}
		else if(name=="\\")	{
			if(stackDepth==0)
				edgeData._operators.push_back(ROLL_LEFT);
			else
				branchData._operators.push_back(ROLL_LEFT);
		}
		else if(name=="/")	{
			if(stackDepth==0)
				edgeData._operators.push_back(ROLL_RIGHT);
			else
				branchData._operators.push_back(ROLL_RIGHT);
		}
		else if(name=="~")	{
			if(stackDepth==0)
				edgeData._operators.push_back(CHANGE_DIRECTION);
			else
				branchData._operators.push_back(CHANGE_DIRECTION);
		}
		//Weight is used to set the relative probabilities for the productions
		else if(name=="Weight")
			weightFound = i;
		else if(i - 1 == weightFound)	//Weight is previous word
			succData._weight = atoi( name.c_str() );
		//Condition
		else if(name==IF)
			conditionFound = i;
		else if(i - 2 == conditionFound && conditionFound!=-1)	{	//If is previous word
			if(name==GREATER_THAN)
				succData._conditionData._condition = COUNTER_GREATER_THAN;
			else if(name==LESSER_THAN)
				succData._conditionData._condition = COUNTER_LESS_THAN;
			else if(name==EQUAL)
				succData._conditionData._condition = COUNTER_EQUAL;
		}
		else if(i - 3 == conditionFound && conditionFound!=-1)	//The constant to compare
			succData._conditionData._value = atoi( name.c_str() );
	}
	if(weightFound==-1)	//Weight not found
		succData._weight = 1;
	return succData;
}

//Rewrite rule is the genr8 name for a production, ie a row in the grammar

void Scaffold::parseRewriteRule(Sentence* rewrite)
{
	vector<int> succTypes, branchStart;
	vector<Branch*> branches;
	int predPos, endPos = rewrite->numberOfWords();
	Sentence *pred, *succ;
	predPos = rewrite->findWord("->");
	//cout << "-> has position " << predPos << endl;
	if(predPos!=0)	{	//Non-Probabalistic rule
		pred = rewrite->getSentence(0, predPos);
		//Get the data about the predecessor
		_predecessorData.push_back( parsePredecessor(pred) );
		//Now parse the successor
		succ = rewrite->getSentence(predPos+1, endPos);
		//Get all the segments that we should transform
		_successorData.push_back( parseSuccessor(succ) );
	}
	else	{	//This is one of several possible successors to a predecessor higher up.
		_predecessorData[_predecessorData.size() - 1]._nSucc++;
		//Now parse the successor
		succ = rewrite->getSentence(predPos+1, endPos);
		//Get all the segments that we should transform
		_successorData.push_back( parseSuccessor(succ) );
	}
}

PredecessorData Scaffold::parsePredecessor(Sentence* pred)
{
	PredecessorData predData;
	EdgeTypeAndCount edge;
	predData.clear();
	edge.clear();
	int nWords = pred->numberOfWords();
	if(nWords==1)	{	//No context sensitivity
		pred->getWord(0)->getEdgeTypeAndCount(edge);
		predData._rewrite = edge;
	}
	//Right or left context-sensitive?
	else if(nWords==3)	{
		pred->getWord(0)->getEdgeTypeAndCount(edge);
		if(pred->getWord(1)->getName()=="<")	{
			predData._prev = edge;
			pred->getWord(2)->getEdgeTypeAndCount(edge);
			predData._rewrite = edge;
		}
		else	{
			predData._rewrite = edge;
			pred->getWord(2)->getEdgeTypeAndCount(edge);
			predData._foll = edge;
		}
	}
	else if(nWords==5)	{	//Right and left context-sensitive
		pred->getWord(0)->getEdgeTypeAndCount(edge);
		predData._prev = edge;
		pred->getWord(2)->getEdgeTypeAndCount(edge);
		predData._rewrite = edge;
		pred->getWord(4)->getEdgeTypeAndCount(edge);
		predData._foll = edge;
	}
	//Store the data
	return predData;
}

void Scaffold::parseAngle(Sentence* sent)
{
	_angle = atoi( sent->getWord(1)->getName().c_str() )*PI / 180;
}

////////////////////////////////////////////
//
//	Saving and loading
//
//	This is not working currently.
//	The strategy is to save all the data (member variables) of the scaffold, vertices, segments and regions to a file.
//	When we read the file, we will get the same scaffold.
//	In between, we can change the environment, giving us a dynamic environment.
//	I have added a few more variables since I wrote these functions, so one would have to go through them and make sure that it is all in there.
//	I suspect that there is not much left to do before it works (unfortunately, I do not have the time).
//
///////////////////////////////////////////////

void Scaffold::saveScaffold(ofstream &file)
{
	cout << "Saving grammar for scaffold " << _ID << endl;
	//Simply write all the data to the file
	//ID
	file << _ID << endl;
	file << _nVertices << endl;
	file << _nSegments << endl;
	file << _nRegions << endl;
	file << _nCurves << endl;
	file << _centerPos.getX() << endl << _centerPos.getY() << endl << _centerPos.getZ() << endl;
	file << _syncronous << endl;
	file << _maxX << endl << _minX << endl << _maxY << endl << _minY << endl << _maxZ << endl << _minZ << endl;
	file << _fixedPerimeter << endl;
	file << _fixedCenter << endl;
	file << _fixedCenterPos.getX() << endl << _fixedCenterPos.getY() << endl << _fixedCenterPos.getZ() << endl;
	file << _pointConst << endl;
	file << _pointExponent << endl;
	file << _nSteps << endl;
	file << _angle << endl;
	//file << _parsed << endl;
	file << _branchLength << endl;
	file << _branchAngle << endl;
	cout << "Saving vertices." << endl;
	//Save the data for each vertex
	for(int i=0; i<_nVertices; i++)
		_vertices[i]->saveVertex(file);
	cout << "Saving segments." << endl;
	//Save the segment data
	for(int j=0; j<_nSegments; j++)
		_segments[j]->saveSegment(file);
	cout << "Saving regions." << endl;
	//Save the regions data
	for(int k=0; k<_nRegions; k++)
		_region[k]->saveRegion(file);
	cout << "Saving curves." << endl;
	//Save the curve data
	for(int l=0; l<_nCurves; l++)	{
		file << _curves[l]->getID() << endl;
		file << _verticesInCurve[l] << endl;
	}
}

void Scaffold::loadScaffold(ifstream &file)
{
	cout << "Loading scaffold." << endl;
	int ID;
	char buf[80];

	file.getline(buf, 80);
	ID = atoi(buf);
	file.getline(buf, 80);
	_nVertices = atoi(buf);
	createVertices();
	file.getline(buf, 80);
	_nSegments = atoi(buf);
	createSegments();
	file.getline(buf, 80);
	_nRegions = atoi(buf);
	createRegions();
	file.getline(buf, 80);
	_nCurves = atoi(buf);
	file.getline(buf, 80);
	_centerPos.setX( atof(buf) );
	file.getline(buf, 80);
	_centerPos.setY( atof(buf) );
	file.getline(buf, 80);
	_centerPos.setZ( atof(buf) );
	file.getline(buf, 80);
	_syncronous = atoi(buf)!=0;
	file.getline(buf, 80);
	_maxX = atoi(buf);
	file.getline(buf, 80);
	_minX = atoi(buf);
	file.getline(buf, 80);
	_maxY = atoi(buf);
	file.getline(buf, 80);
	_minY = atoi(buf);
	file.getline(buf, 80);
	_maxZ = atoi(buf);
	file.getline(buf, 80);
	_minZ = atoi(buf);
	file.getline(buf, 80);
	_fixedPerimeter = atoi(buf)!=0;
	file.getline(buf, 80);
	_fixedCenter = atoi(buf)!=0;
	file.getline(buf, 80);
	_fixedCenterPos.setX( atof(buf) );
	file.getline(buf, 80);
	_fixedCenterPos.setY( atof(buf) );
	file.getline(buf, 80);
	_fixedCenterPos.setZ( atof(buf) );
	file.getline(buf, 80);
	_pointConst = atof(buf);
	file.getline(buf, 80);
	_pointExponent = atoi(buf);
	file.getline(buf, 80);
	_nSteps = atoi(buf);
	file.getline(buf, 80);
	_angle = atof(buf);
	file.getline(buf, 80);
	_branchLength = atof(buf);
	file.getline(buf, 80);
	_branchAngle = atof(buf);
	cout << "Loading vertices" << endl;
	for(int i=0; i<_nVertices; i++)
		_vertices[i]->loadVertex(file, _segments, _region);
	cout << "Loading segments" << endl;
	//Load the segment data
	for(int j=0; j<_nSegments; j++)	{
		_segments[j]->loadSegment(file, _vertices, _region);
	}
	cout << "Loading regions" << endl;
	//Load the regions data
	for(int k=0; k<_nRegions; k++)
		_region[k]->loadRegion(file, _segments, _vertices);
	cout << "Loading " << _nCurves << " curves" << endl;
	//Load the curve data
	file.getline(buf, 80);
	while(!file.eof())	{
		file.getline(buf, 80);
		ID = atoi(buf);
		//cout << "Curve begins with segment " << ID << " and buf contains " << buf << endl;
		_curves.push_back(_segments[ID]);
		file.getline(buf, 80);
		ID = atoi(buf);
		//cout << "Curve has " << ID << " vertices." << endl;
		_verticesInCurve.push_back(ID);
	}
	/*
	//Read data in the same order that they were written
	file >> _ID;
	file >> _nVertices;
	cout << "There are " << _nVertices << " vertices." << endl;
	createVertices();
	file >> _nSegments;
	createSegments();
	file >> _nRegions;
	createRegions();
	file >> x;
	cout << "x is " << x << endl;
	file >> y;
	file >> z;
	_centerPos.setVector(x, y, z);
	file >> _syncronous;
	file >> _maxX;
	file >> _minX;
	file >> _maxY;
	file >> _minY;
	file >> _maxZ;
	file >> _minZ;
	file >> _fixedPerimeter;
	file >> _fixedCenter;
	file >> x;
	file >> y;
	file >> z;
	_fixedCenterPos.setVector(x, y, z);
	file >> _pointConst;
	file >> _pointExponent;
	file >> _nSteps;
	file >> _angle;
	file >> _turnAngle;
	file >> _pitchAngle;
	file >> _rollAngle;
	file >> _branchRange;
	cout << "Loading vertices" << endl;
	for(int i=0; i<_nVertices; i++)
		_vertices[i]->loadVertex(file, _segments, _region);
	cout << "Loading segments" << endl;
	//Load the segment data
	for(int j=0; j<_nSegments; j++)	{
		if(file.rdstate()==ios::goodbit)
			cout << "good" << endl;
		if(file.rdstate()==ios::eofbit)
			cout << END_VAR << endl;
		if(file.rdstate()==ios::failbit)
			cout << "fail" << endl;
		if(file.rdstate()==ios::badbit)
			cout << "bad" << endl;
		if(!file.good())
			file.clear();
		_segments[j]->loadSegment(file, _vertices, _region);
	}
	cout << "Loading regions" << endl;
	if(file.rdstate()==ios::goodbit)
		cout << "good" << endl;
	if(file.rdstate()==ios::eofbit)
		cout << END_VAR << endl;
	if(file.rdstate()==ios::failbit)
		cout << "fail" << endl;
	if(file.rdstate()==ios::badbit)
		cout << "bad" << endl;
	if(!file.good())
		file.clear();
	//file.clear();
	//file >> ID;
	//cout << "Next in file is " << ID << endl;
	//Load the regions data
	for(int k=0; k<_nRegions; k++)
		_region[k]->loadRegion(file, _segments, _vertices);
	cout << "Loading curves" << endl;
	//Load the curve data
	for(int l=0; l<_nCurves; l++)	{
		file >> ID;
		cout << "Curve begins with segment " << ID << endl;
		_curves.push_back(_segments[ID]);
	}*/
}