//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Scaffold.h: interface for the Scaffold class.
//
//	This is a big class that handles the scaffold, ie the internal model of the surface.
//	It is basically a graph data structure, but we need to keep track of more info than an ordinary graph, 
//	because it is in 3D and because of the operations that we need to perform.
//
//	The grammar operates on the segments, so we must keep track of them to be able to apply the productions
//	However, for movment in space, we want to use points, ie the vertices.
//	Thus we move around the vertices as a point cloud and keep track of how they are connected using the segments
//	We also use the vertices as edit-points (or CVs depending on how) when we draw the surface
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SCAFFOLD_H__46203A31_1718_11D5_95D7_00C04F797C24__INCLUDED_)
#define AFX_SCAFFOLD_H__46203A31_1718_11D5_95D7_00C04F797C24__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning(disable:4786)

#define ANGLE							90
#define BRANCH_ANGLE					30

#define START_LENGTH 1
#define START_DIR_X 1
#define START_DIR_Y 0
#define START_DIR_Z 0
#define PI 3.1415926535897932384626433

//How the surface should be drawn in Maya
typedef enum {SURFACE, BSPLINES, POLYGONS, QUILT} surfaceRepresentation;

#include "Vector.h"
#include "Branch.h"
#include "Segment.h"
#include "Region.h"
#include "Vertex.h"
#include "OperatorType.h"
#include <vector>
#include <string>
#include <algorithm>

class World;
class RewriteRule;
class Axiom;
class Predecessor;
class LSystem;
class Sentence;

using namespace std;

typedef enum {NO_CONDITION, COUNTER_GREATER_THAN, COUNTER_LESS_THAN, COUNTER_EQUAL} ConditionType;

//These structs are used to store the information from the grammar when it is being parsed by the scaffold

//Store info about the conditions for a production
struct ConditionData
{
	//What type of condition
	ConditionType			_condition;
	int						_value;	//The required value for the counter of the segment
	void clear()			{_condition = NO_CONDITION; _value = -1;};
};

struct PredecessorData
{
	EdgeTypeAndCount		_prev, _foll, _rewrite;	//We can context sensitivity
	int						_nSucc;	//The number of segments that the predecessor will be split into
	void clear()			{_nSucc = 1; _prev.clear(); _foll.clear(); _rewrite.clear();};
};

//Info about that the branches that shall be placed
struct BranchData
{
	int						_type;
	vector<OperatorTypes>	_operators;	//The turns before we determine the orientation of the branch
	void					clear()	{_type = -1; _operators.clear();};
};

//The edges that the predecessor is split into
struct EdgeData
{
	vector<OperatorTypes>	_operators;		//The turns before placing the segment
	vector<BranchData>		_branches;
	EdgeTypeAndCount		_type;
	void clear()			{_operators.clear(); _type.clear(); _branches.clear();};
};

struct SuccessorData
{
	vector<EdgeData>		_edges;
	int						_weight;
	ConditionData			_conditionData;
	void					clear()	{_edges.clear(); _weight = 1; _conditionData.clear();};
};


class Scaffold  
{
public:
	Scaffold();
	Scaffold(World* world, int ID, bool fixedPerimeter, bool fixedCenter, double pc, int pe, double limitedBranches, int branchAngle, bool softBoundary, int randRewrite, bool fullRandomRewrite, double branchLength, double angle = PI/2);
	virtual ~Scaffold();
	
	void	clear();
	int		getID()	{return _ID;};

	//Position
	bool	getCenter(Vector &pos);
	double	getMaxX()	{return _maxX;};
	double	getMinX()	{return _minX;};
	double	getMaxY()	{return _maxY;};
	double	getMinY()	{return _minY;};
	double	getMaxZ()	{return _maxZ;};
	double	getMinZ()	{return _minZ;};
	//Parameters
	int		getAngle()	{return _angle*180/PI;};
	void	setAngle(int angle)	{_angle = angle*PI/180;};
	void	setSyncronous(bool sync)	{_syncronous = sync;};
	void	setBranchAngle(int branchAngle)	{_branchAngle = branchAngle*PI/180;};
	int		getBranchAngle()	{return _branchAngle*180/PI;};
	void	setNumberOfEdgeTypes(int types)	{_nEdgeTypes = types;};
	bool	isSyncronous()	{return _syncronous;};
	//Vertices
	Vertex*	getVertex(int i)	{return _vertices[i];};
	int		numberOfVertices()	{return _nVertices;};
	//Segments
	int		numberOfSegments()	{return _nSegments;};
	//Set the seed
	void	setSeed(Vector startPos, double length);
	void	setSeed(vector<Vector> cvs, vector<double> masses, int nEdgeTypes);
	//Drawings
	void	drawScaffold(surfaceRepresentation surfRep);
	void	eraseDrawings();
	//Growing
	void	step(double scaleX, double scaleY, double scaleZ);
	void	parseSentence(Sentence* sentence);
	int		numberOfSteps()	{return _nSteps;};
	bool	uniformRegions(double minDifference, double relMinDifference);
	//Loading and saving vertices does not work right now, but it should not be too much work fixing it
	void	saveScaffold(ofstream &file);
	void	loadScaffold(ifstream &file);

	//Mostly for debugging
	void	printSegments();
	void	printCurve(int i);
	void	printVertices();
	void	printRegions();
private:
	void	moveCenter(Vector pos);	//Move center of scaffold to pos
	//Methods
	void	stepFinished(int maxScale);			//Restores the rewritten flags
	void	geometricCenter();
	//Parsing
	PredecessorData		parsePredecessor(Sentence* pred);
	SuccessorData		parseSuccessor(Sentence* succ);
	void				parseRewriteRule(Sentence* rewrite);
	void				parseAngle(Sentence* sent);
	//Growing
	void				segmentsToRewrite(vector<Segment*> &segs, PredecessorData predData);
	void				rewrite(vector<Segment*> &preds, vector<SuccessorData> succDatas);
	void				splitSegments(vector<Segment*> &splits, Segment* preds, int parts);	//Get the split segments in the vector
	void				joinBranches();
	void				addNeighbours(Region* reg, vector<Region*> &neigbs);
	void				splitRegion(Segment* &divider, Region* oldReg, Vertex* start, Vertex* end);
	void				repellPoints();
	void				growScaffold(double scaleX, double scaleY, double scaleZ);
	void				getSuccessors(int predecessor, vector<SuccessorData> &succDatas);
	SuccessorData		getRandomSuccessor(vector<SuccessorData> succDatas);
	
	//Constructors for the parts
	void	newSegment(Segment* &newSegment, Segment &oldSeg);
	void	newSegment(Segment* &newSegment);
	void	newSegment(Segment* &newSegment, Vertex* start, Vertex* end, Region* reg);
	void	newVertex(Vertex* &newVertex, bool perimeter);
	void	newVertex(Vertex* &newVertex, int curve, bool outside, bool perimeter);
	void	newRegion(Region* &newRegion);
	void	newCurve(Segment* newCurve);
	//These are used when we load from file
	void	createVertices();
	void	createSegments();
	void	createRegions();
	
	bool	nextVertex(int curve, Vertex* &current, Segment* &segment);	//Returns true if there are more vertices for the curve
	//Goes to the next segment with even or odd type
	//Returns true if it could go to the next segment
	bool	nextVertex(Vertex* &current, Segment* &segment, bool oddType);
	bool	curveEnds(int curve);
	//Used when we want to create a real nurbs surface
	vector<Vertex*>		updateCVArray();
	
	//Data
	//The parts constituting the scaffold, store them as a simple array
	int _ID;
	vector<Vertex*>		_vertices;
	int					_nVertices;
	vector<Region*>		_region;
	int					_nRegions;
	vector<Segment*>	_segments;
	int					_nSegments;
	//Curves is used to keep track on how we should draw our lines.
	//It is optimized for speed, if we wanted it to look better, we should use a better
	//algorithm that finds longer sequences of vertices to put in one curve.
	vector<Segment*>	_curves;
	int					_nCurves;
	vector<int>			_verticesInCurve;
	//This is used when we want to create a real nurbs-surface
	int					_uKnots, _vKnots;
	//We need a pointer to the world to get info from it
	World*				_world;
	//Position
	Vector				_centerPos;
	double				_maxX, _minX, _maxY, _minY, _maxZ, _minZ;
	//Some growth parameters
	bool				_fixedPerimeter;	//If true, the perimeter will not move
	bool				_fixedCenter;		//If true, the center will always be located
	Vector				_fixedCenterPos;	//in the position of this vector
	//For repelling points
	double				_pointConst;
	int					_pointExponent;
	int					_nSteps;	//the number of steps grown
	//Parameters
	bool				_syncronous;
	double				_angle;		//The angle for the turtle to turn
	//More parameters
	bool				_softBoundary;	//Allow the scaffold to grow through the walls
	bool				_fullRandomRewrite; //Selection of production is fully randomized, otherwise we take the same production for each segment of a given type when using stochastic productions
	int					_randRewrite;	//the noise for the productions
	//Parsing
	bool				_parsed;	//True when we have parsed the grammar
	vector<PredecessorData>	_predecessorData;	//The result of the grammar parsing
	vector<SuccessorData>	_successorData;		//can be found in these two vectors
	Sentence*			_axiom;					//The axiom is parsed by the setSeed function
	int					_nEdgeTypes;	//The number of different edgetypes in the grammar
	
	vector<string*>		_curveNames;			//Keep track of the Maya names of the objects that we have drawn so that we can delete them later
	//Branches
	double				_branchLength, _branchAngle, _characteristicLength;	//Branch info
	int					_nBranches;	//Reset after each step, the total number of branches in the entire scaffold

};

#endif // !defined(AFX_SCAFFOLD_H__46203A31_1718_11D5_95D7_00C04F797C24__INCLUDED_)
