//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Segment.cpp: implementation of the Segment class.
//
//////////////////////////////////////////////////////////////////////

#include "Region.h"
#include <iostream>
#include "Segment.h"
#include "Region.h"
#include "Scaffold.h"

#include "math.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Segment::Segment()
{

}

Segment::Segment(int ID, bool perimeter, int count)
{
	_ID = ID;

	_type = 0;
	_perimeter = perimeter;
	_leftRegion = NULL;
	_rightRegion = NULL;
	_end = NULL;
	_start = NULL;
	_length = 0;
	_curve = 0;
	_rewritten = false;
	_age = 0;
	_counter = count;
}

Segment::Segment(int ID, const Segment &oldSeg)
{
	_ID = ID;
	
	_dir = oldSeg._dir;
	_leftRegion = oldSeg._leftRegion;
	_rightRegion = oldSeg._rightRegion;
	_end = oldSeg._end;
	_start = oldSeg._start;
	_length = oldSeg._length;
	_type = oldSeg._type;
	_perimeter = oldSeg._perimeter;
	_curve = oldSeg._curve;
	_rewritten = oldSeg._rewritten;
	_age = 0;
	_counter = oldSeg._counter;
}

Segment::~Segment()
{

}

//Used to change the type of the segment.
//We can have a noise level of randRewrite which might lead to a random type instead of the intended.

inline void Segment::setType(int type, int randRewrite, int nTypes)
{
	_type = rand() % 100 < randRewrite ? rand() % nTypes : type;
}

void Segment::setOrientation()
{
	_dir = _end->getPosition() - _start->getPosition();
	_length = _dir.normalize();
}

void Segment::switchDirection()
{
	//Swap start and end
	Vertex *tmpVer;
	tmpVer = _start;
	_start = _end;
	_end = tmpVer;
	_dir *= -1;
	//Swap the left and right
	Region *tmpReg;
	tmpReg = _leftRegion;
	_leftRegion = _rightRegion;
	_rightRegion = tmpReg;
}

//This function is used when we turn the segment
//We get a sequence of operators in ops, an angle and the starting direction in &dir

double Segment::changeOrientation(vector<OperatorTypes> ops, double angle, Vector &dir)
{
	//cout << "Changing orientation of segment " << _ID << " with length " << _length << endl;
	bool switchDir = false;
	for(int i=0; i<ops.size(); i++)	{
		//cout << "Found operator " << ops[i] << " with angle " << angle << " for segment " << _ID << endl;
		if(ops[i]==TURN_RIGHT)
			dir.turn(angle);
		else if(ops[i]==TURN_LEFT)
			dir.turn(-angle);
		else if(ops[i]==PITCH_DOWN)
			dir.pitch(angle);
		else if(ops[i]==PITCH_UP)
			dir.pitch(-angle);
		else if(ops[i]==ROLL_RIGHT)
			dir.roll(angle);
		else if(ops[i]==ROLL_LEFT)
			dir.roll(-angle);
		else if(ops[i]==CHANGE_DIRECTION)
			switchDir = !switchDir;
	}
	dir.normalize();
	_dir = dir;
	
	if(switchDir)
		switchDirection();
	
	return _length;
}

//This function is used to adjust the segments after a split and rewrite so that all the new
//segments have the same length. It is recursive (I gave up on the loop) and it is called by
//rewrite for the first original segment. newLength is the length for all the segments and
//depth is the number of segments that we split the old one into.

void Segment::changeLength(double newLength, int depth)
{
	Vertex* ver;
	Segment* seg = this;
	_length = newLength;
	depth--;
	_end->setPosition(_start->getPosition() + _dir*_length);
	//Go to the next segment recursively
	if(depth==0)	{
		return;
	}
	else	{
		//The segments automatically belong to the same curve.
		_end->getNextOnCurve(_curve, ver, seg);
		seg->changeLength(newLength, depth);
	}
}

//Returns the segments connecting to _end (exluding itself)

void Segment::getFollowingSegs(vector<Segment*> &segs)
{
	segs.clear();
	int nSegs = _end->numberOfSegments();
	for(int i=0; i<nSegs; i++)
		segs.push_back(_end->getSegment(i));
}

//Returns the segments connecting to _start (exluding itself)

void Segment::getPreviousSegs(vector<Segment*> &segs)
{
	segs.clear();
	int nSegs = _start->numberOfSegments();
	for(int i=0; i<nSegs; i++)
		segs.push_back(_start->getSegment(i));
}

//Used to see if the segment has the correct context

bool Segment::isTypeFollowing(int type)
{
	int nSegs = _end->numberOfSegments();
	for(int i=0; i<nSegs; i++)	{
		if(_end->getSegment(i)->getType()==type && _end->getSegment(i)->getID()!=_ID)
			return true;
	}
	return false;
}

//Check that we have the rightside-context

bool Segment::isTypeAndCountFollowing(EdgeTypeAndCount edge)
{
	int nSegs = _end->numberOfSegments();
	for(int i=0; i<nSegs; i++)	{
		if(_end->getSegment(i)->getType()==edge._type && _end->getSegment(i)->getID()!=_ID)	{
			if( (edge._counterType==NUMERIC && _counter==edge._counter) || edge._counterType!=NUMERIC )
			return true;
		}
	}
	return false;
}

//Check that we ahve the leftside-context

bool Segment::isTypePrevious(int type)
{
	int nSegs = _start->numberOfSegments();
	for(int i=0; i<nSegs; i++)	{
		if(_start->getSegment(i)->getType()==type && _start->getSegment(i)->getID()!=_ID)
			return true;
	}
	return false;
}

bool Segment::isTypeAndCountPrevious(EdgeTypeAndCount edge)
{
	int nSegs = _start->numberOfSegments();
	for(int i=0; i<nSegs; i++)	{
		if(_start->getSegment(i)->getType()==edge._type && _start->getSegment(i)->getID()!=_ID)	{
			if( (edge._counterType==NUMERIC && _counter==edge._counter) || edge._counterType!=NUMERIC )
			return true;
		}
	}
	return false;
}

//This function checks if we are allowed to apply the production to a specific segment because of the Kondition involved

bool Segment::checkCounter(EdgeTypeAndCount edgeTAndC)
{
	//Check the counter of the segment vs the rewrite-rule
	if( edgeTAndC._counterType==NUMERIC )	{
		//cout << "Checking counter, it has the value " << _counter << " and the desired is " << edgeTAndC._counter << endl;
		if(_counter!=edgeTAndC._counter)
			return false;
	}
	//cout << "Returning true" << endl;
	return true;
}

//This function checks if a segment can be rewritten because of the Kondition

bool Segment::checkCondition(ConditionData condData)
{
	//cout << "Checking condition, counter is " << _counter << ", condition is " << condData._condition << " and the number is " << condData._value << endl;
	//Check the counter against the condition
	if(condData._condition==COUNTER_GREATER_THAN)	{
		if(condData._value >= _counter)
			return false;
	}
	else if(condData._condition==COUNTER_LESS_THAN)	{
		if(condData._value <= _counter)
			return false;
	}
	else if(condData._condition==COUNTER_EQUAL)	{
		if(condData._value != _counter)
			return false;
	}
	//cout << "Returning true" << endl;
	return true;
}

///////////////////////////////////////////////
//
//	Saving and loading
//
///////////////////////////////////////////////

//This will probably have to be updated

void Segment::saveSegment(ofstream &file)
{
	//file << _ID << endl;
	file << _type << endl;
	file << _curve << endl;
	file << _start->getID() << endl;
	file << _end->getID() << endl;
	if(_rightRegion!=NULL)
		file << _rightRegion->getID() << endl;
	else
		file << -1 << endl;
	if(_leftRegion!=NULL)
		file << _leftRegion->getID() << endl;
	else
		file << -1 << endl;
	file << _perimeter << endl;
	file << _dir.getX() << " " << _dir.getY() << " " << _dir.getZ() << endl;
	file << _length << endl;
}

void Segment::loadSegment(ifstream &file, vector<Vertex*> &vers, vector<Region*> &regs)
{
	double x, y, z;
	int ID;
	if(file.fail())
		file.clear();
	file >> _type;
	file >> _curve;
	file >> ID;
	_start = vers[ID];
	file >> ID;
	_end = vers[ID];
	file >> ID;
	if(ID!=-1)
		_rightRegion = regs[ID];
	else
		_rightRegion = NULL;
	file >> ID;
	if(ID!=-1)
		_leftRegion = regs[ID];
	else
		_leftRegion = NULL;
	file >> _perimeter;
	file >> x;
	file >> y;
	file >> z;
	_dir.setVector(x, y, z);
	file >> _length;
}