//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Vector.cpp: implementation of the Vector class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include "Vector.h"
#include "math.h"
#include <sstream>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Vector::Vector()
{
	_x = 0;
	_y = 0;
	_z = 0;
	_normalized = false;
	_length = 0;
}

inline Vector::Vector(double x, double y, double z)
{
	_x = x;
	_y = y;
	_z = z;
	_length = sqrt(_x*_x + _y*_y + _z*_z);
	_normalized = _length==1;
}

Vector::Vector(const Vector& vec)
{
	_x = vec._x;
	_y = vec._y;
	_z = vec._z;
	_length = vec._length;
	_normalized = vec._normalized;
}

Vector::~Vector()
{

}

const bool Vector::operator==(const Vector &vec)
{
	return ( _x==vec._x && _y==vec._y && _z==vec._z );
}

const Vector Vector::operator+(const Vector &vec)
{
	register Vector ret(_x+vec._x, _y+vec._y, _z+vec._z);
	ret.calculateLength();
	return ret;
}

const Vector& Vector::operator +=(const Vector &vec)
{
	_x += vec._x;
	_y += vec._y;
	_z += vec._z;
	calculateLength();
	return *this;
}

const Vector& Vector::operator=(const Vector &vec)
{
	_x = vec._x;
	_y = vec._y;
	_z = vec._z;
	_normalized = vec._normalized;
	_length = vec._length;
	return *this;
}

const Vector& Vector::operator/=(const double &denominator)
{
	if(denominator!=0)	{
		_x /= denominator;
		_y /= denominator;
		_z /= denominator;
		_length /= denominator;
	}
	return *this;
}

const Vector Vector::operator/(const double &denominator)
{
	register Vector ret;
	if(denominator!=0)
		ret.setVector(_x/denominator, _y/denominator, _z/denominator);

	ret.calculateLength();
	return ret;
}

Vector Vector::operator-(const Vector &vec)
{
	register Vector ret(_x-vec._x, _y-vec._y, _z-vec._z);
	ret.calculateLength();
	return ret;
}

const Vector& Vector::operator -=(const Vector &vec)
{
	_x -= vec._x;
	_y -= vec._y;
	_z -= vec._z;
	calculateLength();
	return *this;
}

const Vector& Vector::operator*=(const double &factor)
{
	_x *= factor;
	_y *= factor;
	_z *= factor;
	_length *= factor;
	_normalized = (_length == 1);
	return *this;
}

const Vector Vector::operator*(const double &factor)
{
	register Vector ret(_x*factor, _y*factor, _z*factor);
	ret.calculateLength();
	return ret;
}

const double Vector::operator*(const Vector &vec)
{
	register double prod = 0;
	prod += _x*vec._x;
	prod += _y*vec._y;
	prod += _z*vec._z;
	return prod;
}

const Vector Vector::operator&(const Vector &vec)
{
	register Vector ret;
	ret.setX( _y*vec._z - vec._y*_z );
	ret.setY( _x*vec._z - vec._x*_z );
	ret.setZ( _x*vec._y - vec._x*_y );
	ret.calculateLength();
	return ret;
}

double Vector::normalize()
{
	if(!_normalized)	{
		_length = sqrt(_x*_x + _y*_y + _z*_z);
		if(_length != 0)	{
			_x /= _length;
			_y /= _length;
			_z /= _length;
			_normalized = true;
		}
	}
	return _length;
}

inline void Vector::setLength(double length)
{
	normalize();
	_x *= length;
	_y *= length;
	_z *= length;
	_length = length;
}

void Vector::turn(double angle)
{
	if(!isNormalized())
		normalize();
	//Matrix multiplication
	register double x, y, s, c;
	x = _x;
	y = _y;
	s = sin(angle);
	c = cos(angle);
	
	_x = c*x + s*y;
	_y = -s*x + c*y;
}

void Vector::pitch(double angle)
{
	if(!isNormalized())
		normalize();
	//Matrix multiplication
	register double x, z, s, c;
	x = _x;
	z = _z;
	s = sin(angle);
	c = cos(angle);

	_x = c*x - s*z;
	_z = s*x + c*z;
}

void Vector::roll(double angle)
{
	if(!isNormalized())
		normalize();
	//Matrix multiplication
	register double z, y, s, c;
	z = _z;
	y = _y;
	s = sin(angle);
	c = cos(angle);

	_y= c*y - s*z;
	_z = s*y + c*z;
}

string Vector::printVector()
{
	ostringstream print;
	print << _x << ", " << _y << ", " << _z;
	return print.str().c_str();
}

inline const double Vector::maxv()
{
	register double m = MAXV(_x, _y);
	return MAXV(m, _z);
}

inline const double Vector::minv()
{
	register double m = MINV(_x, _y);
	return MINV(m, _z);
}

void Vector::randv(bool norm)
{
	_x = rand() - RAND_MAX/2;
	_y = rand() - RAND_MAX/2;
	_z = rand() - RAND_MAX/2;
	if(norm)
		normalize();
}

inline void Vector::calculateLength()
{
	_length = sqrt(_x*_x + _y*_y + _z*_z);
	_normalized = _length==1;
}