//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Vertex.h: interface for the Vertex class.
//
//	This class is used for the nodes in the scaffold class. They keep track of their data and the neighbourhood.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_VERTEX_H__46203A34_1718_11D5_95D7_00C04F797C24__INCLUDED_)
#define AFX_VERTEX_H__46203A34_1718_11D5_95D7_00C04F797C24__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning(disable:4786)

#include "Segment.h"
#include "OperatorType.h"
#include "Vector.h"
#include <vector>
#include <fstream>
#include "Field.h"

class Segment;
class Boundary;
class Force;
class Noise;
class Region;
class Branch;
struct BranchData;

using namespace std;

class Vertex : public Field	//Since they can be repelling, we inherit the force-properties
{
public:
	Vertex();
	Vertex(int ID, Boundary* boundary, Force* force, Noise* noise, double limitedBranches, double branchAngle, bool outside, bool perimeter, double mass = 1);
	Vertex(Vector pos);
	virtual ~Vertex();

	//Scale proportionally when growing
	void	scale(double factor, Vector center);
	void	scale(double factorX, double factorY, double factorZ, Vector center, bool softBoundary = false);
	Vector	getPosition()	{return _pos;};
	void	setPosition(Vector pos)	{_pos = pos;};
	void	moveVertex(Vector move, bool walls = false);	//Move the vertex with or without the effect of the walls
	int		getID()	{return _ID;};
	double	getMass()	{return _mass;};
	void	setMass(double mass)	{_mass = mass;};
	//The vertex knows what segments surround it
	void	addSegment(Segment* seg);
	bool	removeSegment(Segment* seg);
	//And the regions too..
	void	addRegion(Region* reg);
	bool	removeRegion(Region* reg);
	bool	regionAdjacent(Region* reg);
	void	adjacentRegions(vector<Region*> &regions);
	Region*	getRegion(int i)	{return _regions[i];};
	int		numberOfRegions()	{return _regions.size();};
	//Get info about the curves and the neigbourhood
	bool	curveEnds(int curve, Segment* &seg);	//Returns true if there are two segments that belong to curve
	bool	getNextOnCurve(int curve, Vertex* &ver, Segment* &seg);
	int		numberOfSegments()	{return _segments.size();};
	Segment*	getSegment(int i)	{return _segments[i];};
	bool	adjacent(Vertex* ver);	//Returns true if ver is adjacent
	void	adjacentVertices(vector<Vertex*> &vertices);
	bool	hasCurve(int curve);
	void	curvesPassing(vector<int> &curves);
	double	moveToCenterOfNeighbours();
	Vector	getDisplacement(bool noise = true);	//See how it is affected by forces
	bool	onPerimeter()	{return _perimeter;};
	void	setPerimeter(bool perimeter = true)	{_perimeter = perimeter;};
	Vector	getForce(Vector pos, double konst, int exp);
	bool	isOutside()	{return _outside;};
	void	setOutside(bool outside)	{_outside = outside;};
	//Branches
	void	placeBranch(BranchData branchData, Vector dir, double angle, int branchID);
	int		numberOfBranches()	{return _branches.size();};
	Branch*	getBranch(int i)	{return _branches[i];};
	bool	matchBranch(Branch* branch, int &type, Region* &reg, bool &start, Vector center, double characteristicLength);
	void	removeBranch(Branch* branch);
	void	removeAllBranches();
	//To use when we save the scaffold
	void	saveVertex(ofstream &file);
	void	loadVertex(ifstream &file, vector<Segment*> &segs, vector<Region*> &regs);

	void	printInfo();
	
private:
	//Methods
	double		directionAndDifference(Vector otherBranchPos, Vector branchDirection);	//Returns the dot product between the direction of the branch and the difference vector between two branches
	bool		closeEnough(Vertex* otherBranchStart, double characteristicLength);	//Check if the vertices are close enough when connecting branches
	//Data
	int			_ID;
	double		_mass;
	vector<Segment*> _segments;	//The segments that connect to the vertex
	Vector		_pos;
	bool		_perimeter;	//Is the vertex on the perimeter
	vector<Region*>	_regions;//Keep track of the neighboring regions
	vector<Branch*>	_branches;	//Store all the branches
	//The vertex must know where the walls are and have pointers to the environment elements
	Boundary*	_boundary;
	Force*		_force;
	Noise*		_noise;
	double		_branchRange, _branchAngle;
	bool		_outside;	//Has the vertex moved outside a boundary? For use by soft boundary agent.
};

#endif // !defined(AFX_VERTEX_H__46203A34_1718_11D5_95D7_00C04F797C24__INCLUDED_)
