//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// World.cpp: implementation of the World class.
//
//////////////////////////////////////////////////////////////////////

#include "World.h"
#include <sstream>
#include <iostream>
#include <fstream>
#include "AgentSmooth.h"
#include "AgentSubdivision.h"
#include "AgentSize.h"
#include "AgentSymmetry.h"
#include "AgentSoftBoundary.h"
#include "AgentUndulating.h"
#include "MELUtilities.h"
#include <Maya/MComputation.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

World::World()
{

}

World::World(vector<Vector> attrPos, vector<Vector> repPos, bool attrSing, bool repSing, Direction dir, double force, wallBehaviour wallType,  int tiles, double noise, bool fixedPerimeter, bool fixedCenter, double ac, double rc, double wc, int ae, int re, int we, double pc, int pe, int popSize, Fractals fractal, surfaceRepresentation surfRep, double scaleX, double scaleY, double scaleZ, string grammarFile, double limitedBranches, int branchAngle, int tournamentSize, int elites, bool softBoundary, BNFtype bnfType, int genomeLength, int randRewrite, bool fullRandomRewrite, int maxBNFDepth, int maxEdgeTypes, int numberOfScaffolds, bool differentGrammars, int crossoverPoints, bool uniformCells, double branchLength, string encodedGrammar)
{
	//Create the population
	_populationSize = popSize;
	//Data for the EA
	_nElites = popSize > elites ? elites : popSize;
	_tournamentSize = tournamentSize;
	_numberOfScaffolds = numberOfScaffolds;
	_crossoverPoints = crossoverPoints;
	for(int i=0; i<popSize; i++)	{
		Individual* ind = new Individual(i, this, tiles, fixedPerimeter, fixedCenter, pc, pe, fractal, grammarFile, limitedBranches, branchAngle, softBoundary, bnfType, genomeLength, randRewrite, fullRandomRewrite, maxBNFDepth, maxEdgeTypes, numberOfScaffolds, differentGrammars, branchLength, encodedGrammar);
		_individuals.push_back(ind);
	}
	
	_evolution = ( tiles==0 && fractal==NO_FRACTAL && grammarFile=="" && encodedGrammar=="");
	//Drawing parameters
	_surfRep = surfRep;
	_uniformCells = uniformCells;
	//Environment
	_force = new Force(attrPos, repPos, force, dir, ac, rc, ae, re, attrSing, repSing);

	//Create the boundaries
	_boundary = new Boundary(wallType, wc, we);
	
	_noise = new Noise(noise);

	_geneticEngine = new GeneticEngine();

	_scaleX = scaleX;
	_scaleY = scaleY;
	_scaleZ = scaleZ;

	//Agents
	_nAgents = 0;
}

World::~World()
{

}

//Move some parameters from the constructor...
//The agents represent the fitness-function

void World::createAgents(double smooth, double smoothWeight, double ratio, double ratioWeight, double xsize, double ysize, double sizeWeight, double symratio, double symWeight, double softWeight, double undulation, double undulationWeight)
{
	AgentSmooth* agentSmooth = new AgentSmooth(_nAgents++, smoothWeight, smooth);
	_agents.push_back(agentSmooth);
	AgentSubdivision* agentSubdivision = new AgentSubdivision(_nAgents++, ratioWeight, ratio);
	_agents.push_back(agentSubdivision);
	AgentSize* agentSize = new AgentSize(_nAgents++, sizeWeight, xsize, ysize);
	_agents.push_back(agentSize);
	AgentSymmetry* agentSymmetry = new AgentSymmetry(_nAgents++, symWeight, symratio);
	_agents.push_back(agentSymmetry);
	AgentSoftBoundary* agentSoftBoundary = new AgentSoftBoundary(_nAgents++, softWeight);
	_agents.push_back(agentSoftBoundary);
	AgentUndulating* agentUndulating = new AgentUndulating(_nAgents++, undulationWeight, undulation);
	_agents.push_back(agentUndulating);
}

//The individuals are in the _individuals vector is ranked after fitness

Individual* World::getIndividualID(int ID)
{
	for(int i=0; i<_populationSize; i++)	{
		if(_individuals[i]->getID()==ID)
			return _individuals[i];
	}
	return NULL;	//Failed to find the ind
}

int World::getRankingForIndID(int ID)
{
	for(int i=0; i<_populationSize; i++)	{
		if(_individuals[i]->getID()==ID)	{
			//cout << "Individual " << ID << " has rank " << i << endl;
			return i;
		}
	}
	return -1;
}

//Determine fitness for this individual

double World::evaluateIndividual(int ind)
{
	double score = 0;
	//Loop through all the agents and sum the fitness
	for(int i=0; i<_nAgents; i++)
		score += _agents[i]->evaluateScaffold(_individuals[ind]);
	//Impression is the same as fitness
	_individuals[ind]->setImpression(score);
	return score;
}

//Get fitness for the population

void World::evaluatePopulation()
{
	double mean = 0;
	for(int i=0; i<_populationSize; i++)
		mean += evaluateIndividual(i);
	
	_fitnessHistory.push_back(mean/_populationSize);
}

void World::evaluatePopulation(vector<double> fitness)
{
	double mean = 0;
	for(int i=0; i<_populationSize; i++)	{
		if(fitness[i]==-1)
			mean += evaluateIndividual(i);
		else	{
			mean += fitness[i];
			_individuals[i]->setImpression(fitness[i]);
		}
	}
	_fitnessHistory.push_back(mean/_populationSize);
}

void World::generateGrammars()
{
	for(int i=0; i<_populationSize; i++)	{
		//_individuals[i]->getGenome()->printGenome();
		//cout << "Generating grammar for individual " << _individuals[i]->getID() << endl;
		_individuals[i]->generateGrammar();
	}
}

void World::interpretGrammars()
{
	for(int i=0; i<_populationSize; i++)	{
		//cout << "Interpreting grammar for individual " << _individuals[i]->getID() << endl;
		_individuals[i]->interpretGrammar();
	}
	//cout << "Grammars interpreted" << endl;
}

void World::rankPopulation()
{
	sort(_individuals.begin(), _individuals.end(), CompareFitness());
}

void World::clear()
{
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->clear();
}

void World::eraseDrawings()
{
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->eraseScaffold();
}

void World::draw(int viewInd, int step)
{
	for(int i=0; i<_populationSize; i++)	{
		cout << "Drawing individual " << _individuals[i]->getID() << " with fitness " << _individuals[i]->getImpression() << endl;
		if(viewInd!=-1)
			_individuals[i]->drawScaffold(_surfRep, viewInd==i);
		else
			_individuals[i]->drawScaffold(_surfRep, true, step);
	}
	if(_evolution)	{
		//cout << "Mean fitness is: ";
		//for(int j=_fitnessHistory.size()-1; j>0; j--)
		//	cout << _fitnessHistory[j] << ", ";
		//cout << _fitnessHistory[_fitnessHistory.size()-1] << endl;
	}
	refresh();
}

void World::growScaffold()
{
	for(int j=0; j<_populationSize; j++)	{
		//cout << "Growing scaffold for individual " << _individuals[j]->getID() << endl;
		_individuals[j]->step(_scaleX, _scaleY, _scaleZ);
		if(_uniformCells)
			_individuals[j]->makeCellsUniform(0.01, 0);
	}
}

//Create the scaffolds for the population

bool World::generateScaffolds(int steps, bool drawScaffold, int viewInd)
{
	MComputation comp;
	comp.beginComputation();
	for(int i=0; i<steps; i++)	{
		growScaffold();
		//cout << "Scaffold grown." << endl;
		if(drawScaffold)	{
			if(_populationSize==1)	{
				_individuals[0]->setVisibility(false);
				draw(viewInd, i + 1);
			}
			else	{
				eraseDrawings();
				draw(viewInd);
			}
		}
		if(comp.isInterruptRequested())	{
			cout << "Interrupted!" << endl;
			if(!drawScaffold)
				draw(viewInd);
			return false;
		}
	}
	comp.endComputation();
	//cout << "Scaffolds generated." << endl;
	return true;
}

//Use this for user-defined seeds

void World::setSeed(vector<Vector> seed, vector<double> masses, bool drawScaffold, int viewInd)
{
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->setSeed(seed, masses);
	if(drawScaffold)	{
		if(_populationSize==1)
			draw(viewInd, 0);
		else
			draw(viewInd);
	}
}

//Use this for the genetically defined seeds

void World::setSeed(bool drawScaffold, int viewInd)
{
	for(int i=0; i<_populationSize; i++)	{
		for(int j=0; j<_startPositions.size(); j++)	{
			//cout << "Setting seed for scaffold " << j << endl;
			//cout << "Startlength is " << _startLengths[j] << " and pos is " << _startPositions[j].printVector().c_str() << endl;
			_individuals[i]->setSeed(_startPositions[j], _startLengths[j], j);
		}
	}
	if(drawScaffold)	{
		if(_populationSize==1)
			draw(viewInd, 0);
		else
			draw(viewInd);
	}
}

void World::breedPopulation()
{
	//Take out the elite members that should be transfered directly into the new population
	//Make copies of the other genomes
	vector<Genotype*> geneCopies;
	Genotype *newGenome, *newGenome2;
	for(int i=0; i<_populationSize; i++)	{
		//cout << "Copying the genome of individual " << _individuals[i]->getID() << endl;
		geneCopies.push_back(_individuals[i]->getGenome());
	}
	int mother, father;
	for(int j=_nElites; j<_populationSize; j++)	{
		newGenome = new Genotype(0);
		newGenome2 = new Genotype(0);
		mother = getRankingForIndID(tournamentSelect());
		father = getRankingForIndID(tournamentSelect());
		_geneticEngine->discreteCrossover(geneCopies[mother], geneCopies[father], newGenome, newGenome2, _crossoverPoints);
		_geneticEngine->mutate(newGenome);
		_individuals[j]->setGenome(newGenome);
		delete newGenome2;
	}
	for(int k=_nElites; k<_populationSize; k++)
		delete geneCopies[k];
	//cout << "Deleted old genome." << endl;
}

int World::tournamentSelect()
{
	//Take a few individuals
	vector<Individual*> _participants;
	for(int i=0; i<_tournamentSize; i++)
		_participants.push_back(_individuals[rand() % _populationSize]);
	//Rank them
	sort(_participants.begin(), _participants.end(), CompareFitness());
	//Return the winner
	return _participants[0]->getID();
}
/*
//Use this when we load scaffolds (needs updating)
void World::initialize(string fileName, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness)
{
	ifstream file;
	if(!opengenr8File(fileName, file))
		return;
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->getScaffold()->loadScaffold(file);
	file.close();
	if(cont)
		loadGenomes(loadGenome);
	//Interpret the genomes as grammars
	generateGrammars();
	interpretGrammars();
	generateScaffolds(steps, drawEachStep);
	if(_evolution)	{
		//Evaluate the scaffolds
		evaluatePopulation(fitness);
		//Rank them
		rankPopulation();
	}
	if(!drawEachStep)
		draw();
}
*/

//For user-defined seeds

bool World::initialize(vector<double> masses, vector<Vector> seed, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness, int viewInd, int copyInd, int nCopy, bool identicalCopies)
{
	ifstream file;
	_userDefinedSeed = true;
	_seed = seed;
	_masses = masses;
	//cout << "Continue is " << boolalpha << cont << " and loadGenomeFile is " << loadGenome.c_str() << endl;
	if(cont || loadGenome!="genome.txt")
		loadGenomes(loadGenome, !cont);
	//Make copies of a genome
	copyGenome(copyInd, nCopy, identicalCopies);
	//Interpret the genomes as grammars
	generateGrammars();
	interpretGrammars();
	setSeed(seed, masses, drawEachStep, viewInd);
	if(!generateScaffolds(steps, drawEachStep, viewInd))
		return false;
	if(_evolution)	{
		//Evaluate the scaffolds
		evaluatePopulation(fitness);
		//Rank them
		rankPopulation();
	}
	//Draw the stuff
	if(!drawEachStep)
		draw(viewInd);
	return true;
}

bool World::initialize(vector<Vector> startPos, vector<double> startLength, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness, int viewInd, int copyInd, int nCopy, bool identicalCopies)
{
	ifstream file;
	_userDefinedSeed = false;
	for(int j=0; j<_numberOfScaffolds; j++)	{
		_startLengths.push_back(startLength[j]);
		_startPositions.push_back(startPos[j]);
	}
	if(cont || loadGenome!="genome.txt")
		loadGenomes(loadGenome, !cont);
	//Make copies of a genome
	copyGenome(copyInd, nCopy, identicalCopies);
	//Interpret the genomes as grammars
	generateGrammars();
	interpretGrammars();
	setSeed(drawEachStep, viewInd);
	if(!generateScaffolds(steps, drawEachStep, viewInd))
		return false;
	if(_evolution)	{
		//Evaluate the scaffolds
		evaluatePopulation(fitness);
		//Rank them
		rankPopulation();
	}
	//Draw the stuff
	if(!drawEachStep)
		draw(viewInd);
	return true;
}

bool World::step(int steps, bool drawEachStep, int viewInd)
{
	//Clean up old stuff
	eraseDrawings();
	clear();
	//Mix the genetic material a little
	//if(_evolution)	{
		breedPopulation();
		//Interpret the genomes as grammars
		generateGrammars();
		//Construct scaffolds from the grammars
		interpretGrammars();
	//}
	//Place the seed
	if(_userDefinedSeed)
		setSeed(_seed, _masses, drawEachStep, viewInd);
	else
		setSeed(drawEachStep, viewInd);
	//Generate the scaffolds
	if(!generateScaffolds(steps, drawEachStep, viewInd))
		return false;
	if(_evolution)	{
		//Evaluate the scaffolds
		evaluatePopulation();
		//Rank them
		rankPopulation();
	}
	//Draw the stuff
	if(!drawEachStep)	{
		//eraseDrawings();
		draw(viewInd);
	}
	return true;
}

//Use this to make copies fo an inidividual in the population

void World::copyGenome(int copyInd, int nCopy, bool identicalCopies)
{
	//cout << "nCopy is " << nCopy << " and copyInd is " << copyInd << endl;
	if(nCopy>0)	{
		int start = _populationSize - nCopy;
		//Make sure that we get the right number of copies
		start = start <= copyInd ? start-1 : start;
		//Bound check
		start = start < 0 ? 0 : start;
		//cout << "Start: " << start << endl;
		Genotype *copy, *original = _individuals[copyInd]->getGenome();
		//Put the copies at the shallow end of the genepool (at the end of the pop)
		for(int i=start; i<_populationSize; i++)	{
			if(i!=copyInd)	{
				delete _individuals[i]->getGenome();
				copy = new Genotype(original);
				if(!identicalCopies)
					_geneticEngine->mutate(copy);
				_individuals[i]->setGenome(copy);
			}
		}
	}
}

void World::saveGrammar(string fileName)
{
	ofstream file;
	if(!opengenr8File(fileName, file))	{
		cout << "Failed to open " << fileName << endl;
		return;
	}
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->saveGrammar(file);
	
	cout << "Closing file." << endl;
	file.close();
	cout << "Done saving." << endl;
}

void World::saveScaffolds(string fileName)
{
	ofstream file;
	if(!opengenr8File(fileName, file))
		return;
	for(int i=0; i<_populationSize; i++)
		_individuals[i]->getScaffold()->saveScaffold(file);

	cout << "Closing file." << endl;
	file.close();
	cout << "Done saving." << endl;
}

void World::saveGenomes(string fileName)
{
	ofstream file;
	if(!opengenr8File(fileName, file))
		return;
	cout << "Trying to save genomes in file " << fileName << endl;
	file << _populationSize << endl;
	for(int i=0; i<_populationSize; i++)	{
		if(_individuals[i]->getGenome())
			_individuals[i]->getGenome()->saveGenome(file);
	}

	file.close();
}

void World::loadGenomes(string fileName, bool fullPath)
{
	int popsize;
	ifstream file;
	cout << "Loading genomes" << endl;
	if(!opengenr8File(fileName, file, fullPath, fullPath))
		return;
	
	file >> popsize;
	popsize = popsize <= _populationSize ? popsize : _populationSize;
	//Put the new individuals in the shallow end of the genepool
	for(int i=_populationSize - popsize; i<_populationSize; i++)
		_individuals[i]->getGenome()->loadGenome(file);

	file.close();
}