//
// Copyright (C) 2001 Martin Hemberg. You are allowed to use or modify the code, but this note must be kept intact.
// 
// World.h: interface for the World class.
//
//	The World contains all the relevant data about the GENR8 environment.
//	It contains the forces, the boundaries and the population.
//	It is created by the as genr8 is started, but at that point it hardly holds any information.
//	Constructor only creates a minimum of objects, most things are instead created by initialize or during the main loop.
//	This because we sometimes need to create World objects even though we are not going to do a full run (eg regn8)
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_WORLD_H__4DC59882_1B08_11D5_95D7_00C04F797C24__INCLUDED_)
#define AFX_WORLD_H__4DC59882_1B08_11D5_95D7_00C04F797C24__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Individual.h"
#include "Force.h"
#include "Boundary.h"
#include "Noise.h"
#include "GeneticEngine.h"
#include "Agent.h"
#include <vector>
#include <algorithm>

using namespace std;

class Scaffold;

//Use this to rank the population with the built in sort.
class CompareFitness {
public:
	int operator() (Individual* ind1, Individual* ind2)
	{return (ind1->getImpression() < ind2->getImpression());}
};

class World  
{
public:
	World();
	World(vector<Vector> attrPos, vector<Vector> repPos, bool attrSing, bool repSing, Direction dir, double force, wallBehaviour wallType, int tiles, double noise, bool fixedPerimeter, bool fixedCenter, double ac, double rc, double wc, int ae, int re, int we, double pc, int pe, int popSize, Fractals fractal, surfaceRepresentation surfRep, double scaleX, double scaleY, double scaleZ, string grammarFile, double limitedBranches, int branchAngle, int tournamentSize, int elites, bool softBoundary, BNFtype bnfType, int genomeLength, int randRewrite, bool fullRandomRewrite, int maxBNFDepth, int maxEdgeTypes, int numberOfScaffolds, bool differentGrammars, int crossoverPoints, bool uniformCells, double branchLength, string encodedGrammar);
	virtual ~World();

	//Use this in addition to the constructor (so that we do not have to have all arguments there)
	void		createAgents(double smooth, double smoothWeight, double ratio, double ratioWeight, double xsize, double ysize, double sizeWeight, double symratio, double symWeight, double softWeight, double undulation, double undulationWeight);
	//Access to a lot of stuff
	Force*		getForce()		{return _force;};
	Scaffold*	getScaffold(int i=0)	{return _individuals[i]->getScaffold();};
	Boundary*	getBoundary()	{return _boundary;};
	Noise*		getNoise()		{return _noise;};
	//Get info about the population, individuals tend to be rather mixed after a while.
	Individual*	getIndividualRanked(int i)	{return _individuals[i];};
	Individual*	getIndividualID(int i);
	int			getRankingForIndID(int i);
	int			getPopulationSize()	{return _populationSize;};
	GeneticEngine*	getGeneticEngine()	{return _geneticEngine;};
	//Genesis
	void		initialize(string fileName, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness);
	bool		initialize(vector<double> masses, vector<Vector> seed, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness, int viewInd, int copyInd, int nCopy, bool identicalCopies);
	bool		initialize(vector<Vector> startPos, vector<double> length, int steps, bool drawEachStep, bool cont, string loadGenome, vector<double> fitness, int viewInd, int copyInd, int nCopy, bool identicalCopies);
	//One generation
	bool		step(int steps, bool drawEachStep = true, int viewInd = -1);	//Main loop
	void		saveScaffolds(string fileName);
	void		saveGrammar(string fileName);
	void		saveGenomes(string fileName);
private:
	//Methods
	void		rankPopulation();
	//Set the fitness for the pop
	void		evaluatePopulation();
	//Use this to override the fitness function with user-defined values
	void		evaluatePopulation(vector<double> fitness);
	double		evaluateIndividual(int i);
	//Mix the genepool
	void		breedPopulation();
	int			tournamentSelect();
	//Interpret the genomes
	void		generateGrammars();
	//Let the scaffolds parse the grammars
	void		interpretGrammars();
	void		clear();
	void		eraseDrawings();
	void		draw(int viewInd, int step = -1);
	//Grow the scaffolds
	bool		generateScaffolds(int steps, bool draw, int viewInd);	//Return false if computation was interrupted
	void		setSeed(vector<Vector> seed, vector<double> masses, bool drawScaffold, int viewInd);
	void		setSeed(bool drawScaffold, int viewInd);
	void		growScaffold();
	void		makeCellsUniform();
	void		loadGenomes(string fileName, bool fullPath);
	void		copyGenome(int copyInd, int nCopy, bool identicalCopies);
	//Here we should have population.....
	vector<Individual*>	_individuals;
	int			_populationSize;
	int			_genomeLength, _nElites, _tournamentSize;
	vector<double>	_fitnessHistory;
	//These are the same for all individuals
	//Environmental elements
	Force*		_force;
	Boundary*	_boundary;
	Noise*		_noise;
	GeneticEngine*	_geneticEngine;
	vector<Agent*>	_agents;	//Fitness
	int			_nAgents, _numberOfScaffolds, _crossoverPoints;
	bool		_userDefinedSeed, _evolution, _uniformCells;
	surfaceRepresentation	_surfRep;	//Drawing style
	vector<Vector>	_seed, _startPositions;
	vector<double>	_startLengths, _masses;
	double			_scaleX, _scaleY, _scaleZ;
};

#endif // !defined(AFX_WORLD_H__4DC59882_1B08_11D5_95D7_00C04F797C24__INCLUDED_)
